import sys
from fractions import Fraction
from inspect import signature, Signature, Parameter
from types import UnionType
from typing import *

import rich.traceback

from ..Func import *
from ...data_structures import Heap
from ..LiteralEval import literal_eval

overloads = {}  # type: dict[str, Heap]
overload_temp = {}  # type: dict[str, dict[tuple[type, ...], OverloadFunctionCallable]]
overload_funcs = {}  # type: dict[str, OverloadFunctionCallable]


# TODO: 完成模块

def _register(qual_name):
    if qual_name not in overload_funcs:
        overload_funcs[qual_name] = OverloadFunctionCallable(qual_name)
    return _get_registered(qual_name)


def _get_registered(qual_name):
    return overload_funcs[qual_name]


def _add_to_temp(qual_name, types):
    if qual_name not in overload_temp:
        overload_temp[qual_name] = dict()
    overload_temp[qual_name][types] = _get_registered(qual_name)


def _check_temp(qual_name, args):
    if qual_name not in overload_temp:
        return False
    for types in overload_temp[qual_name]:
        if len(types) != len(args):
            continue
        for arg, type_ in zip(args, types):
            if not isinstance(arg, type_):
                return False


def _match_type(arg, param_type):
    if isinstance(param_type, Parameter):
        param_type = param_type.annotation

    if isinstance(param_type, str):
        param_type = literal_eval(param_type, globals=globals(), locals=locals(), builtins=True, no_eval=False)

    origin = get_origin(param_type)

    if origin is Union:
        types = get_args(param_type)
        matches = [_match_type(arg, t) for t in types]
        return sum(matches) / len(types)

    elif origin in (List, Set, Deque, FrozenSet, Sequence):  # TODO: Not finished
        inner_types = get_args(param_type)
        if not inner_types:
            return 1  # No specific type specified, assume match
        inner_type = inner_types[0]
        if isinstance(arg, origin):  # 如果arg是序列类型之一
            return sum(_match_type(item, inner_type) for item in arg) / len(arg) if arg else 1
        else:
            return 0

    elif origin is Tuple:
        inner_types = get_args(param_type)
        if not inner_types:
            return 1  # No specific type specified, assume match
        if isinstance(arg, tuple):  # 如果arg是元组
            return sum(_match_type(item, inner_type) for item, inner_type in zip(arg, inner_types)) / len(
                arg) if arg else 1
        else:
            return 0

    elif origin is Dict:
        kt, vt = get_args(param_type)
        if isinstance(arg, dict):  # 如果arg是字典
            return sum(
                (_match_type(v, vt) + _match_type(k, kt)) / 2
                for k, v in arg.items()) / len(arg) if arg else 1
        else:
            return 0
    else:
        return 1 if isinstance(arg, param_type) else 0


def _get_match_degree(signature: Signature, args):
    match_degrees = []
    bound_args = signature.bind(*args)
    for param_name, arg in bound_args.arguments.items():
        param = signature.parameters[param_name]
        match_degrees.append(_match_type(arg, param.annotation))

    return sum(match_degrees) / len(match_degrees) if match_degrees else 0  # 匹配程度


def count_possible_types(type_hint):
    # if type_hint is type:
    #     return 1
    print("Type:", type_hint)
    origin = get_origin(type_hint)
    if origin in (Union, UnionType):
        # 如果是Union类型，递归计算每个成员的可能类型数量
        return sum(count_possible_types(arg) for arg in get_args(type_hint))
    elif origin is List:
        # 如果是List类型，递归计算元素类型的可能类型数量
        element_type = get_args(type_hint)[0]
        return count_possible_types(element_type)
    elif origin is Tuple:
        # 如果是Tuple类型，递归计算每个元素类型的可能类型数量
        return sum(count_possible_types(arg) for arg in get_args(type_hint))
    elif origin is Optional:
        # 如果是Optional类型，递归计算内部类型的可能类型数量，并加上None
        return count_possible_types(get_args(type_hint)[0]) + 1
    else:
        # 基本类型或其他类型，返回1
        return 1


class ErrorDecorater:
    def __init__(self, exc):
        self.exception = exc

    def __bool__(self):
        return False

    def __str__(self):
        return str(self.exception)

    __repr__ = __str__


class OverloadError(Exception):
    def __init__(self, qual_name, tests=(), args=(), kwargs=None):
        self.qual_name = qual_name
        self.tests = tests
        self.msg = ''
        self.args = args
        self.kwargs = kwargs if kwargs else {}

    @staticmethod
    def to_call_format(args, kwargs):
        string = ', '.join(
            list(map(lambda x: type(x).__name__, args)) +
            list(map(lambda x: f'{x[0]}={type(x[1]).__name__}', kwargs._instances()))
        )
        return string

    @staticmethod
    def to_args_format(signature: Signature):
        params = signature.parameters
        string = ', '.join(
            list(map(str, params.values()))
        )
        return string

    @staticmethod
    def to_type_name(type_: inspect.Parameter):
        if type_.annotation is inspect.Parameter.empty:
            return 'Any'
        else:
            return str(type_.annotation.__name__)

    @staticmethod
    def __rich_to_string(string):
        console = rich.get_console()
        rich_string = console.render(string)
        result = ''
        for part in rich_string:
            result += str(part.text)
        return result

    def __generate_error_msg(self):
        error_msg = ''

        def add_error_msg(string):
            nonlocal error_msg
            error_msg += self.__rich_to_string(string)

        add_error_msg(f'[red]调用{self.qual_name}时发生错误: 无法匹配重载类型[/red]')
        add_error_msg(f'[yellow]传入的实参:[/yellow] {self.to_call_format(self.args, self.kwargs)}')
        for test, func in zip(self.tests, overloads[self.qual_name]):
            add_error_msg(
                f'\t[yellow]尝试匹配: {func.signature} [/yellow]'
                f'[red]: {str(test)}[/red]'
            )
        return error_msg

    def __str__(self):
        try:
            return self.__generate_error_msg()
        except Exception as e:
            return str(e)


class OverloadRuntimeError(Exception):
    def __init__(self, qualname, called_overload, e, call_args, call_kwargs):
        self.qualname = qualname
        self.called_overload = called_overload
        self.e = e
        self.call_args = call_args
        self.call_kwargs = call_kwargs

    def __str__(self):
        return (f'\n\t以参数({OverloadError.to_call_format(self.call_args, self.call_kwargs)})'
                f'调用`{self.qualname}`'
                f'的重载 "{self.called_overload}" '
                f'时发生错误.'
                f'\n\t详细信息 {self.e.__class__.__name__}: {self.e}'
                )


def _get_module_globals(module_name):
    try:
        module = sys.modules[module_name]
        return module.__dict__
    except KeyError:
        return {}


class OverloadFunction:
    def __init__(self, func):
        self.func = func
        self.qual_name = get_qualname(func)
        self.signature = signature(func)
        self.params = dict(self.signature.parameters)
        self.prec = self.get_prec(self.signature)
        # print("重载精确度:", self.prec)

    @staticmethod
    def get_prec(signature: Signature):
        return Fraction(
            sum(count_possible_types(param.annotation) for param in signature.parameters.values()),
            len(signature.parameters))

    def __lt__(self, other):
        return self.prec < other.prec

    def __eq__(self, other):
        return self.prec == other.prec

    def __gt__(self, other):
        return self.prec > other.prec

    def __str__(self):
        return f'Overload({self.signature}) with prec {self.prec}'

    __repr__ = __str__


class OverloadFunctionCallable:
    def __init__(self, qualname):
        self.qualname = qualname

    def _call(self, func: OverloadFunction, args, kwargs):
        try:
            return func.func(*args, **kwargs)
        except Exception as e:
            raise OverloadRuntimeError(self.qualname, func.signature, e, args, kwargs)

    def __call__(self, *args, **kwargs):
        if _check_temp(self.qualname, args):
            return
        res_ls = []
        func: OverloadFunction
        for func in overloads[self.qualname]:
            res = _get_match_degree(func.signature, args)
            res_ls.append((res, func))

        res, matched_func = max(res_ls)
        if res < 1:
            raise OverloadError(self.qualname, tuple(res_ls), args, kwargs)

        return self._call(matched_func, args, kwargs)


def overload(func):
    func = OverloadFunction(func)
    if func.qual_name in overloads:
        overloads[func.qual_name].append(func)
    else:
        overloads[func.qual_name] = Heap([func], True)

    return _register(func.qual_name, func)


def get_func_overloads(func: OverloadFunctionCallable):
    return overloads[func.qualname]



import argparse
import os
import subprocess
import sys
import time

import rich.traceback
from rich import print

rich.traceback.install()


def run_command(command):
    ps = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return ps.returncode, ps


def init_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--install", '-i',
        help="Install HydrogenLib",
        action="store_true"
    )
    parser.add_argument(
        "--skip-check", '-c',
        help="Skip check HydrogenLib wheel",
        action="store_true"
    )
    parser.add_argument(
        "--skip-upload", '-d',
        help="Skip upload HydrogenLib wheel",
        action="store_true"
    )
    parser.add_argument(
        '--skip-build', '-s',
        help="Skip building HydrogenLib wheel",
        action="store_true"
    )
    parser.add_argument(
        "--clean", '-k',
        help="Clean HydrogenLib wheel",
        action="store_true"
    )
    parser.add_argument(  # 接收一个参数，version
        "--version", '-v',
        help="Set HydrogenLib version",
        default="None",
        type=str
    )


version_path = r"src/hydrolib/src\version"
args = sys.argv[1::]

if __name__ == '__main__':
    spinner = "aesthetic"
    parser = argparse.ArgumentParser()
    console = rich.console.Console(force_terminal=True)
    init_parser(parser)
    args = parser.parse_args(args)
    if args.version != "None":
        with console.status("Setting version...", spinner=spinner):
            rt_code, ps = run_command(["hatch", "version", args.version])
        time.sleep(0.1)
        if rt_code != 0:
            console.print("[bold red]Setting version failed!")
            console.print(ps.stderr.decode())
            console.print(ps.stdout.decode())
            sys.exit(rt_code)
        print("[bold green]success!")
    if args.clean:
        if os.name == 'nt':
            command = ["powershell.exe", "-Command", "rm", r".\dist\*"]
        elif os.name == 'posix':
            command = ["rm", "-rf", r"./dist/*"]
        else:
            console.print(f"[bold red]Unsupported OS({os.name})!")
            sys.exit(1)

        with console.status("Cleaning old files...", spinner=spinner):
            rt_code, ps = run_command(command)
        time.sleep(0.1)
        if rt_code != 0:
            console.print("[bold red]Cleaning old files failed!")
            console.print(ps.stderr.decode())
            console.print(ps.stdout.decode())
            sys.exit(rt_code)
        print("[bold green]success!")

    if not args.skip_build:
        # 播放工作动画
        with console.status("Building wheel...", spinner=spinner):
            rt_code, ps = run_command(["hatch", "build"])
        time.sleep(0.1)
        if rt_code != 0:
            console.print("[bold red]Building wheel failed!")
            console.print(ps.stderr.decode())
            sys.exit(rt_code)
        print("[bold green]success!")
    # console.console.print('\n')
    if not args.skip_check:
        with console.status("Checking wheel...", spinner=spinner):
            rt_code, ps = run_command(["twine", "check", "dist/*"])
        time.sleep(0.1)
        if rt_code != 0:
            console.print("[bold red]Checkin wheel failed!")
            console.print(ps.stderr)
            sys.exit(rt_code)
        print("[bold green]success!")
    # console.console.print('\n')
    if not args.skip_upload:
        print('Upload...')
        # ps = subprocess.run(["twine", "upload", r".\dist\*", '--disable-progress-bar'])
        ps = subprocess.run(["twine", "upload", r".\dist\*"])
        rt_code = ps.returncode
        if rt_code != 0:
            console.print(f"[bold red]failed!({rt_code})")
            console.print(ps.stdout.decode())
            console.print(ps.stderr.decode())
            sys.exit(rt_code)
        print("[bold green]success!")

    print(
        "[green]All steps are success, "
        "you can run "
        "[/green]`[#FFA500]pip[/#FFA500] install [blue]HydrogenLib[#FFA500]-[/#FFA500]Next[/blue] "
        "[dim]-U[/dim]` "
        "[green]to update HydrogenLib.")



from .Interpreter import Lexer, Parser, Interpreter


def get_lexer(source_code):
    return Lexer.lex(source_code)


def get_parser(tokens):
    return Parser.parser(tokens)


def get_interpreter(ast):
    it = Interpreter.Interpreter(ast)
    return it


def get(source_code):
    lex = get_lexer(source_code)
    ast = get_parser(lex)
    return get_interpreter(ast)




# Namedpipe for linux and macos
# TODO: 完成基于FIFO的命名管道伪实现的测试
import fcntl
import os
import time
from queue import Queue
from threading import Event

from ..._hycore.threading_methods import run_with_timeout, run_new_daemon_thread


# module end, but test not finished

def create_fifo(path):
    try:
        os.mkfifo(path)
    except FileExistsError:
        return


def lock_fifo(fd):
    fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)


def unlock_fifo(fd):
    fcntl.flock(fd, fcntl.LOCK_UN)


class _Reader:
    def __init__(self, name):
        self.name = name
        self._event = Event()
        self._queue = Queue()
        self._thread = run_new_daemon_thread(self._read_thread)

    def _wait(self):
        while not os.path.exists(self.name):
            time.sleep(0.05)

    def wait(self, timeout=None):
        try:
            run_with_timeout(self._wait, timeout)
        except TimeoutError as e:
            raise e

    def _read(self):
        with open(self.name, 'r') as f:
            try:
                lock_fifo(f.fileno())
                data = f.read()
                unlock_fifo(f.fileno())
                return data
            except Exception:
                return None

    def _read_thread(self):
        while not self._event.is_set():
            try:
                data = self._read()
                self._queue.put(data)
            except:
                continue

    def read(self, timeout=None):
        return self._queue.get(timeout=timeout)


class _Writer:
    def __init__(self, name):
        self.name = name
        self._queue = Queue()
        self._event = Event()

        self._thread = run_new_daemon_thread(self._write_thread)

    def check(self):
        if not os.path.exists(self.name):
            create_fifo(self.name)

    def _write(self, data):
        self.check()
        with open(self.name, 'w') as f:
            try:
                lock_fifo(f.fileno())
                f.write(data)
                unlock_fifo(f.fileno())
            except Exception:
                return

    def _write_thread(self):
        while not self._event.is_set():
            try:
                data = self._queue.get(timeout=0.05)
                self._write(data)
            except Exception as e:
                pass

    def write(self, data):
        self._queue.put(data)

    def close(self):
        os.remove(self.name)


def Reader(name):
    return _Reader(name)


def Writer(name):
    return _Writer(name)


if __name__ == '__main__':
    r = Reader("Test")
    w = Writer("Test")
    r.wait()
    w.write("Hello")
    print(r.read())



import asyncio
import socket
from typing import Any, Union


class Asyncsocket:
    """
    socket.socket的异步版本
    """
    def __init__(self, s: Union[socket.socket, Any] = None, event_loop: asyncio.AbstractEventLoop = None):
        if s is None:
            self.sock = socket.socket()
        elif isinstance(s, self.__class__):
            self.sock = s.sock
        else:
            self.sock = s

        if self.sock.getblocking() is True:
            self.sock.setblocking(False)  # 异步IO采用非阻塞

        self.loop = event_loop if event_loop else asyncio.get_running_loop()

    async def sendall(self, data):
        return await self.loop.sock_sendall(
            self.sock, data
        )

    async def recv(self, size: int):
        return await self.loop.sock_recv(
            self.sock, size
        )

    async def recv_into(self, buffer):
        return await self.loop.sock_recv_into(
            self.sock, buffer
        )

    async def accept(self):
        return await self.loop.sock_accept(self.sock)

    async def connect(self, addr, timeout=None):
        if timeout is None:
            return await self.loop.sock_connect(self.sock, addr)
        else:
            return await self.loop.sock_connect(self.sock, addr), timeout

    async def connect_ex(self, addr):
        return self.sock.connect_ex(addr)

    def settimeout(self, timeout=None):
        self.sock.settimeout(timeout)

    def listen(self, backlog):
        self.sock.listen(backlog)

    def detach(self):
        return self.sock.detach()

    def family(self):
        return self.sock.family

    def fileno(self):
        return self.sock.fileno()

    def get_inheriteable(self):
        return self.sock.get_inheritable()

    def getblocking(self):
        return self.sock.getblocking()

    def getpeername(self):
        return self.sock.getpeername()

    def getsockname(self):
        return self.sock.getsockname()

    def getsockopt(self, level, optname, buflen=None):
        if buflen is None:
            return self.sock.getsockopt(level, optname)
        else:
            return self.sock.getsockopt(level, optname, buflen)

    def gettimeout(self):
        return self.sock.gettimeout()

    def ioctl(self, control, option):
        return self.sock.ioctl(control, option)

    async def bind(self, addr):
        self.sock.bind(addr)

    async def close(self):
        self.sock.close()



from setuptools import setup, Extension

module = Extension(
    'hybitmap',
    sources=['hybitmap_capi.c'],
    language='c',
)

setup(
    name='hybitmap',
    version='1.0',
    description='Python C extension for bitmap operations',
    ext_modules=[module],
    python_requires='>=3.8',
    author='LittleSong2024',
    author_email='idesong6@qq.com',
    url='',
    license='HydrogenLib License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ]
)



from ...abc.backend import BackendABC
from ....hystruct import Struct, BinStructBase
from ....hycore import builtin_types


class Binstruct_Backend(BackendABC):
    serializer = Struct()
    support_types = (BinStructBase, builtin_types)

    def save(self):
        with self._fd.open(self.file, 'wb') as f:
            f.write(self.serializer.dumps(self._data))

    def load(self):
        with self._fd.open(self.file, 'rb') as f:
            try:
                if f.size:
                    self.existing = True
                    self.init(**self.serializer.loads(f.read(), mini=True))
            except RuntimeError as e:
                return



from typing import Protocol, Self, Union, runtime_checkable


@runtime_checkable
class Packable_CLS_NO_ARG(Protocol):
    def pack(self, ins) -> bytes:
        ...


@runtime_checkable
class Packable_INS_NO_ARG(Protocol):
    def pack(self) -> bytes:
        ...


@runtime_checkable
class Packable_CLS_WITH_ARG(Protocol):
    @classmethod
    def pack(cls, ins, *args, **kwargs) -> bytes:
        ...


@runtime_checkable
class Packable_INS_WITH_ARG(Protocol):
    def pack(self, *args, **kwargs) -> bytes:
        ...


@runtime_checkable
class Unpackable_CLS_NO_ARG(Protocol):
    @classmethod
    def unpack(cls, data: bytes) -> 'Self':
        ...


@runtime_checkable
class Unpackable_CLS_WITH_ARG(Protocol):
    @classmethod
    def unpack(cls, data: bytes, *args, **kwargs) -> 'Self':
        ...


@runtime_checkable
class Unpackable_INS_NO_ARG(Protocol):
    def unpack(self, data: bytes) -> 'Self':
        ...


@runtime_checkable
class Unpackable_INS_WITH_ARG(Protocol):
    def unpack(self, data: bytes, *args, **kwargs) -> 'Self':
        ...


PackableNoArg = Union[Packable_CLS_NO_ARG, Packable_INS_NO_ARG]
PackableWithArg = Union[Packable_CLS_WITH_ARG, Packable_INS_WITH_ARG]
Packable = Union[PackableNoArg, PackableWithArg]

UnpackableNoArg = Union[Unpackable_CLS_NO_ARG, Unpackable_INS_NO_ARG]
UnpackableWithArg = Union[Unpackable_CLS_WITH_ARG, Unpackable_INS_WITH_ARG]
Unpackable = Union[UnpackableNoArg, UnpackableWithArg]



import socket
from io import *


class SocketIO(IOBase):
    def __init__(self):
        self.data = b''
        self.pos = 0

    def send(self, data):
        self.data += data

    def recv(self, size, flags=0):
        self.pos += size
        data = self.data[self.pos - size:self.pos]
        if flags & socket.MSG_PEEK:
            self.pos -= size

        return data



import os
import platform
import re
import sys

from subprocess import run as _run, PIPE
from argparse import ArgumentParser


def run(*args, **kwargs):
    if kwargs.get('check') is None:
        kwargs['check'] = True
    return _run(*args, **kwargs)


class MyParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.add_argument(
            '--tag', '-t',
            type=str,
            help='tag name',
            required=True
        )

    def get_tag_name(self, args=None):
        if args is None:
            args = sys.argv[1:]

        nsp = self.parse_args(args)
        tag = nsp.tag.removeprefix('refs/tags/')
        return tag


class TagChecker:
    # on:
    #     push:
    #         tags:
    #             - "v-current-upload*"
    #             - "v*.*.*-upload*"
    #             - "v-patch-upload*"
    #             - "v-major-upload*"
    #             - "v-minor-upload*"
    def __init__(self):
        self.re_v_current = re.compile(r'v-(current)-upload.*')
        self.re_v_vvv = re.compile(r'v(\d+\.\d+\.\d+)-upload.*')
        self.re_v_patch = re.compile(r'v-(patch)-upload.*')
        self.re_v_major = re.compile(r'v-(major)-upload.*')
        self.re_v_minor = re.compile(r'v-(minor)-upload.*')

        self.re_ls = [
            self.re_v_current,
            self.re_v_vvv,
            self.re_v_patch,
            self.re_v_major,
            self.re_v_minor,
        ]

    def match(self, tag):
        for re in self.re_ls:
            res = re.match(tag)
            if res:
                return res.group(1)
        return None


class Main:
    def __init__(self):
        self.parser = MyParser()
        self.checker = TagChecker()
        self.version = None
        self.tag = None

    def load_version(self):
        ps = run(['hatch', 'version'], stdout=PIPE)
        version = ps.stdout.decode('utf-8').strip()
        print("Version:", repr(version))
        self.version = version

    def check_tag(self):
        self.tag = tag = self.parser.get_tag_name()
        if tag:
            # 解析标签
            curv = self.checker.match(tag)
            if not curv:
                raise ValueError(f'无法解析标签版本号: {tag}')
            libv = self.version
            # 如果传入的标签版本号小于当前版本号，则不执行上传
            if curv not in {'patch', 'major', 'minor', 'current'} and tuple(curv.split('.')) < tuple(libv.split('.')):
                print('当前版本号大于传入的版本号，不执行上传')
                sys.exit(1)  # 阻止工作流继续执行
            if curv == 'current':
                print('不更新版本号')
            else:
                print('更新版本号:', curv)
                self.update_version(curv)

    def update_version(self, ver):
        run(['hatch', 'version', ver])

    def upload(self):
        run(['twine', 'upload', './dist/*'])

    def build(self):
        run(['hatch', 'build'])

    def exec(self):

        self.load_version()

        print(f'Upload script runs on {platform.platform()}')
        print(f'PID: {os.getpid()}')
        print(f'Args: {sys.argv}')
        print(f'Env: {os.environ}')
        print(f'Activity tag: {self.parser.get_tag_name()}')
        print(f'Get-Version: {self.version}')

        self.check_tag()
        self.build()
        self.upload()


if __name__ == '__main__':
    main = Main()
    main.exec()



from .encrypt.aes import decrypt as aes_decode, encode as aes_encode
from .._hycore.json_file import *
from .._hycore.utils import null


# module end


class _EncryptJsonFile:
    """
    entrypt the value , not key and value
    """

    def __init__(self, filename, key, iv):
        self.iostream = pickle_safe_open(filename, dict)
        self.key, self.iv = key, iv

    def __getitem__(self, item):
        cry_text = self.iostream[item]
        plain_text = aes_decode(cry_text, self.key, self.iv)
        return Pickle.decode(plain_text)

    def __setitem__(self, key, value):
        value = Pickle.encode(value)
        cry_text = aes_encode(value, self.key, self.iv)
        self.iostream[key] = cry_text

    def __delitem__(self, key):
        del self.iostream[key]

    def get(self, key, default=null):
        try:
            x = self.iostream.top(key, default)
            return aes_decode(x, self.key, self.iv)
        except KeyError as e:
            if default is not null:
                return default
            raise e

    def clear(self):
        self.iostream.clear()

    def set(self, key, value):
        self[key] = value

    def keys(self):
        return self.iostream.keys()

    def values(self):
        return self.iostream.values()

    def items(self):
        return self.iostream._instances()

    def __iter__(self):
        return self.iostream.__iter__()

    def save(self):
        self.iostream.save()

    def flush(self):
        self.iostream.flush()

    def close(self):
        self.iostream.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def EncryptJsonOpen(filename, key, iv):
    return _EncryptJsonFile(filename, key, iv)

# TODO: finish this class
# class EncryptSyncSocket(SyncSocket):
#     def __init__(self, s: socket.socket | object | None, aes_key=None, aes_iv=None, rsa_pubkey=None, rsa_prikey=None):
#         super().__init__(s)
#         self._encry_mode = None
#
#         if not any([aes_key, aes_iv, rsa_pubkey, rsa_prikey]):
#             raise ValueError("必须指定aes_key, aes_iv, rsa_pubkey, rsa_prikey")
#
#         if aes_key is not None and aes_iv is not None:
#             self._encry_mode = 'aes'
#             self.key1 = aes_key
#             self.key2 = aes_iv
#
#         elif rsa_pubkey is not None and rsa_prikey is not None:
#             self._encry_mode = 'rsa'
#             self.key1 = rsa_pubkey
#             self.key2 = rsa_prikey
#
#         else:
#             args = [
#                 k for k, v in
#                 {'aes_key': aes_key, 'aes_iv': aes_iv, 'rsa_pubkey': rsa_pubkey, 'rsa_prikey': rsa_prikey}.items() if
#                 v is not None
#             ]
#             raise ValueError(
#                 f"无法确定加密方式(Args: {tuple(args)}"
#             )
#
#     def write(self, data):
#         data = Pickle.encode(data).encode()  # <------------- encode() return bytes
#         if self._encry_mode:
#             if self._encry_mode == 'aes':
#                 encry_data = aes_encode(data, self.key1, self.key2)
#             elif self._encry_mode == 'rsa':
#                 encry_data = rsa_encrypt(data, self.key1)
#             else:
#                 raise ValueError("无效的加密模式")
#             super().write(encry_data)
#
#     def read(self, timeout=None):
#         if self._encry_mode:
#             if self._encry_mode == 'aes':
#                 encry_data = super().read(timeout)
#                 decry_data = aes_decode(encry_data, self.key1, self.key2)
#             elif self._encry_mode == 'rsa':
#                 encry_data = super().read(timeout)
#                 decry_data = rsa_decrypt(encry_data, self.key2)
#             else:
#                 raise ValueError("无效的加密模式")
#
#             return Pickle.decode(decry_data.decode())



import inspect
import types
import typing


def get_args(func):
    for i in inspect.signature(func).parameters.values():
        yield i


def get_name(func):
    return func.__name__


def get_doc(func):
    return func.__doc__


def get_code(func):
    return func.__code__


def get_source(func):
    return inspect.getsource(func)


def get_module(func) -> str:
    return func.__module__


def is_instance(ins_or_cls):
    return not isinstance(ins_or_cls, type)


def get_qualname(func_type_or_ins: typing.Union[types.FunctionType, type, object]):
    if is_instance(func_type_or_ins) and not is_function(func_type_or_ins):
        return get_qualname(func_type_or_ins.__class__)
    return f'{func_type_or_ins.__module__}.{func_type_or_ins.__qualname__}'


def is_function(obj):
    Func_Callable_Types = typing.Union[types.FunctionType, types.BuiltinFunctionType]
    return isinstance(obj, Func_Callable_Types)


class Function:
    def __init__(self, func):
        self._func = func
        self._signature = inspect.signature(func)

    @property
    def name(self):
        return get_name(self._func)

    @property
    def doc(self):
        return get_doc(self._func)

    @property
    def code(self):
        return get_code(self._func)

    @property
    def source(self):
        return get_source(self._func)

    @property
    def module(self):
        return get_module(self._func)

    @property
    def qualname(self):
        return get_qualname(self._func)

    @property
    def signature(self):
        return self._signature

    @property
    def params(self):
        return tuple(get_args(self._func))

    def match(self, *args, **kwargs):
        return self._signature.bind(*args, **kwargs)

    def __str__(self):
        return f'<Func {self.name} {self.qualname}>'

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)


class FunctionGroup:
    def __init__(self, funcs: typing.Iterable):
        self._funcs = funcs

    def __iter__(self):
        return iter(self._funcs)

    def __len__(self):
        return len(self._funcs)

    def __getitem__(self, item):
        return self._funcs[item]

    def __setitem__(self, key, value):
        self._funcs[key] = value

    def __delitem__(self, key):
        del self._funcs[key]

    def __contains__(self, item):
        return item in self._funcs

    def __call__(self, *args, **kwargs):
        for func in self._funcs:
            func(*args, **kwargs)

    def __add__(self, other):
        if isinstance(other, type(self)):
            return type(self)(self._funcs + other._funcs)
        elif isinstance(other, typing.Callable):
            return type(self)(self._funcs + [other])

    def __iadd__(self, other):
        if isinstance(other, type(self)):
            self._funcs += other._funcs
            return self



from threading import Event


class OnceMessage:
    """
    一次性消息
    - set(message, timeout=None)
        设置消息内容,如果上一个消息未被读取,则等待上一个消息被读取
        - message: 消息内容
        - timeout: 等待上一个消息被读取的超时时间,单位为秒,默认为 None
    - get(timeout=None)
        获取消息内容,如果消息未被设置,则等待消息被设置
        - timeout: 等待消息被设置超时时间,单位为秒,默认为 None
    - is_set()
        判断消息是否已被设置
    - clear()
        清除消息,一次性消息将还原为未设置状态
    - close()
        关闭一次性消息,将无法再设置和获取消息
    """

    @property
    def closed(self):
        return self._shutdown

    def __init__(self):
        self._msg = None
        self._sented = Event()
        self._geted = Event()
        self._shutdown = False

    def __del__(self):
        self.close()

    def set(self, message, timeout=None):
        if self._shutdown:
            raise RuntimeError("OnceMessage has been closed")
        self._geted.wait(timeout)  # 等待 get 方法被调用
        self._msg = message
        self._sented.set()  # 通知 get 方法消息已设置
        self._geted.clear()  # 清除 _geted 事件，以便下次使用

    def get(self, timeout=None):
        if self._shutdown:
            raise RuntimeError("OnceMessage has been closed")
        if self._sented.wait(timeout):
            self._sented.clear()  # 清除 _sented 事件，以便下次使用
            self._geted.set()  # 通知 set 方法 get 方法已被调用
            return self._msg
        else:
            raise TimeoutError()

    def is_set(self):
        """
        判断消息是否已被设置
        """
        return self._sented.is_set()

    def clear(self):
        """
        清除消息,一次性消息将还原为未设置状态
        如果不进行新的set操作,上一个消息将会被保留
        """
        self._sented.clear()
        self._geted.clear()

    def close(self):
        self._shutdown = True
        self.clear()



from fractions import Fraction


class ProbabilityCounter:
    def __init__(self, init_value=None):
        if init_value is None:
            init_value = {}
        self.count_dict = dict(init_value)

    def __setitem__(self, key, value):
        self.count_dict[key] = value

    def __getitem__(self, item):
        if item not in self.count_dict:
            self.count_dict[item] = 0
        return self.count_dict[item]

    def get(self, key):
        s = sum(self.count_dict.values())
        if s == 0:
            return Fraction()
        return Fraction(self.count_dict[key], s)

    def update(self, data: dict):
        for key, value in data.items():
            if key not in self.count_dict:
                self.count_dict[key] = 0
            self.count_dict[key] += value

    def increment(self, key, value=1):
        if key not in self.count_dict:
            self.count_dict[key] = 0
        self.count_dict[key] += value

    def probabilities(self, keys=None):
        if keys is None:
            keys = list(self.count_dict.keys())
            keys.sort()
        else:
            keys = list(keys)

        return [self.get(key) for key in keys]

    def proabilities_dict(self, keys=None):
        if keys is None:
            keys = list(self.count_dict.keys())
            keys.sort()
        return dict(zip(keys, self.probabilities(keys)))

    def __iter__(self):
        return iter(self.count_dict)

    def __len__(self):
        return len(self.count_dict)


if __name__ == '__main__':
    counter = ProbabilityCounter()
    for i in ['a', 'b', 'c', 'd']:
        counter[i] += 10

    print(counter.probabilities())



from io import BytesIO

from .data_structures import Stack
from .path import *


def read(file, mode='r'):
    """
    从文件读取数据，允许自定义模式
    """
    with open(file, mode) as f:
        return f.read()


def write(data, file, mode='w'):
    """
    向文件写入数据，允许自定义模式
    """
    with open(file, mode) as f:
        f.write(data)


def empty(file):
    """
    判断文件内容是否为空
    """
    with open(file) as f:
        d = f.read(1)
        if d == '':
            return True


def isspace(file):
    """
    判断文件内容是否为空白
    """
    d = read(file)
    return d.isspace()


fileDataType = Union[str, bytes, bytearray]


class NeoIo:
    """
    新文件读写
    特点:
        1. 可复用
        2. 灵活,完全支持内置的IOAPI
        3. 快速获取文件信息
        4. 堆栈文件

    适用于普通文件操作，在常见桌面和服务器操作系统上表现良好。
    """

    class FState:
        def __init__(self, fstat: os.stat_result):
            self.fstat = fstat

        @property
        def birthday(self):
            """
            创建时间
            可用情况视平台而定，如Windows不可用。
            """
            # 注意：st_birthtime在某些平台上可能不可用
            try:
                return self.fstat.st_birthtime
            except AttributeError:
                return None

        @property
        def last_access(self):
            """
            最后访问时间
            """
            return self.fstat.st_atime

        @property
        def last_modified(self):
            """
            最后修改时间
            """
            return self.fstat.st_mtime

        @property
        def size(self):
            """
            文件大小
            """
            return self.fstat.st_size

        @property
        def mode(self):
            """
            文件权限
            """
            return self.fstat.st_mode

        @property
        def inode(self):
            """
            文件ID
            """
            return self.fstat.st_ino

        @property
        def device(self):
            """
            文件设备
            """
            return self.fstat.st_dev

        @property
        def nlink(self):
            """
            链接数
            """
            return self.fstat.st_nlink

        @property
        def uid(self):
            """
            用户ID
            """
            return self.fstat.st_uid

        @property
        def gid(self):
            """
            组ID
            """
            return self.fstat.st_gid

    def __init__(self):
        self._fd_ls = Stack()
        self.create = False

    def __push_fd(self, fd):
        self._fd_ls.push(fd)

    @property
    def _top_fd(self):
        """
        获取当前栈顶的文件
        """
        return self._fd_ls.at_top

    @classmethod
    def from_fd(cls, fd):
        """
        从文件描述符创建实例
        """
        ins = cls()
        ins.push_new_fd(fd)
        return ins

    @classmethod
    def fopen(cls, file, mode='r', encoding=None, *args, **kwargs):
        """
        打开文件并返回实例
        """
        fd = open(file, mode, encoding=encoding, *args, **kwargs)
        ins = cls.from_fd(fd)
        return ins

    def open(self, file, mode='r', encoding=None, create=None, *args, **kwargs):
        """
        打开一个新的文件,并压入栈中
        """
        if not exists(file) and (self.create if (create is None) else create):
            mkfile(file)
        self.__push_fd(
            open(file, mode, encoding=encoding, *args, **kwargs))
        return self

    def push_new_fd(self, fd):
        """
        压入一个新的文件描述符
        """
        self.__push_fd(fd)

    @property
    def opened(self):
        """
        是否存在打开的文件
        """
        return self._top_fd.at_top is not None and not self._top_fd.closed

    @property
    def can_write(self):
        """
        是否可写
        """
        return self._top_fd.writable()

    @property
    def can_read(self):
        """
        是否可读
        """
        return self._top_fd.readable()

    @property
    def can_seek(self):
        """
        是否可定位
        """
        return self._top_fd.seekable()

    @property
    def is_bytes_io(self):
        """
        是否为BytesIO
        """
        return isinstance(self._top_fd, BytesIO)

    @property
    def pos(self):
        """
        当前位置
        """
        return self._top_fd.tell()

    @property
    def fileno(self):
        """
        文件描述符
        """
        return self._top_fd.fileno()

    @property
    def osfstat(self):
        """
        文件状态
        """
        try:
            return os.fstat(self.fileno)
        except OSError:
            return None

    @property
    def neofstat(self):
        """
        文件状态
        """
        return NeoIo.FState(self.osfstat)

    @property
    def size(self):
        """
        文件大小
        """
        return self.osfstat.st_size

    def write(self, data: Union[bytes, str]):
        """
        写入数据
        """
        if self.can_write:
            self._top_fd.write(data)
        else:
            raise IOError("文件无法写入")

    def seek(self, offset, whence=0):
        """
        定位文件
        """
        self._top_fd.seek(offset, whence)

    def read(self, size=-1):
        """
        读取数据
        """
        if self.can_read:
            return self._top_fd.read(size)
        else:
            raise IOError("文件无法读取")

    def readline(self, size=-1):
        """
        读取一行
        """
        if self.can_read:
            return self._top_fd.readline(size)
        else:
            raise IOError("文件无法读取")

    def readlines(self, hint=-1):
        """
        读取所有行
        """
        if self.can_read:
            return self._top_fd.readlines(hint)
        else:
            raise IOError("文件无法读取")

    def clear(self):
        """
        清除文件内容
        """
        self._top_fd.truncate(0)

    def flush(self):
        """
        刷新文件
        """
        self._top_fd.flush()

    def close(self):
        """
        关闭位于栈顶的文件
        """
        if self._top_fd:
            self._top_fd.close()
            self._fd_ls.pop()
        else:
            raise IOError('未打开任何文件')

    def close_all(self):
        """
        关闭所有文件
        """
        while self._fd_ls:
            self._fd_ls.pop().close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import threading
import time


class TimedData:
    def __init__(self, data, timeout=60, set_time=True, check_time=True, timeout_delete=True):

        self.data = data
        self.time = None
        self._timeout = timeout

        if set_time:
            self.time = time.time()

        self.set_time = set_time
        self.check_time = check_time
        self.timeout_delete = timeout_delete

        self._is_timeout = False  # 缓存

    @property
    def is_timeout(self):
        return self.check()

    @property
    def remain(self):
        if self.check():
            return 0
        return self._timeout - (time.time() - self.time)

    def delete(self):
        """
        Deletes the data.
        """
        if self.data is not None:
            del self.data
            self.data = None

    def _check(self):
        if self.time is None or self.check_time is False:
            # If the time is not set, or if the check is disabled, return True.
            return False

        if time.time() - self.time > self._timeout:
            if self.timeout_delete:
                self.delete()

            return True

        return False

    def check(self):
        if self._is_timeout:
            return self._is_timeout
        self._is_timeout = self._check()
        return self._is_timeout

    def __get__(self, instance, owner):
        if self.check():
            return self.data
        else:
            raise TimeoutError("TimedData has timed out.")


class TimedDataManager:
    def __init__(self, timeout=60):
        self.timeout = timeout
        self.data = {}  # type: dict[str, TimedData]

        self._lock = threading.Lock()

    def check(self, key):
        return self.data[key].check()

    def delete(self, key):
        if self.exists(key):
            self.data[key].delete()
            del self.data[key]

    def delete_threadsafe(self, key):
        with self._lock:
            self.delete(key)

    def delete_multiple(self, keys):
        for key in keys:
            self.delete(key)

    def delete_multiple_threadsafe(self, keys):
        with self._lock:
            self.delete_multiple(keys)

    def extend(self, dic):
        for key, value in dic._instances():
            self.add(key, value)

    def extend_threadsafe(self, dic):
        with self._lock:
            self.extend(dic)

    def get_multiple(self, keys):
        return {key: self.get_data(key) for key in keys if self.exists(key)}

    def get_multiple_threadsafe(self, keys):
        with self._lock:
            return self.get_multiple(keys)

    def add(self, key, value, timeout=None):
        self.data[key] = TimedData(value, timeout=timeout or self.timeout)

    def get(self, key):
        return self.get_data(key).data

    def add_data(self, key, data):
        self.data[key] = data

    def get_data(self, key):
        if not self.check(key):
            return self.data[key]
        else:
            del self.data[key]

    def add_threadsafe(self, key, value, timeout=None):
        with self._lock:
            self.add(key, value, timeout=timeout)

    def get_threadsafe(self, key):
        with self._lock:
            return self.get(key)

    def add_data_threadsafe(self, key, data):
        with self._lock:
            self.add_data(key, data)

    def exists(self, key):
        return key in self.data and not self.check(key)

    def clear(self):
        self.data.clear()

    def clean_up(self):
        results = set()
        for key in filter(self.check, self.data.keys()):
            results.add(key)

        self.delete_multiple_threadsafe(results)



from abc import ABC, abstractmethod

from ._types import *
from ..Lexer import Token

# TODO: 完成递归语法解析
# TODO: 模块暂停开发


class ASTNode(ABC):
    children: list

    @abstractmethod
    def __init__(self, children: list, **kwargs):
        ...


class GeneratorManager:
    def __init__(self, generators: list, next_hook=None):
        self._generators = generators
        self._next_hook = next_hook

    def result(self):
        while True:
            i = 0
            res = False
            if self._next_hook:
                self._next_hook()
            while i < len(self._generators):
                try:
                    res = next(self._generators[i])
                except StopIteration:
                    if isinstance(res, ASTNode):
                        return res
                    else:
                        raise RuntimeError("Unexpected error")
                finally:
                    if res is False:
                        self._generators.pop(i)
                        i -= 1
                i += 1

            if len(self._generators) == 0:
                return None


def _pos_generator_wrap_func(block, index_ls):
    for i in range(len(block.children)):
        if isinstance(block.children[i], Block):
            yield from _pos_generator_wrap_func(block.children[i], index_ls + [i])
        else:
            yield index_ls + [i]


class SyntaxParser:

    def __init__(self):

        self._pos: Pos = None
        self._block: Block = None

        self.end = None
        self._pos_generator = None

    def _pos_generator_func(self):
        return _pos_generator_wrap_func(self._block, [])

    def __getitem__(self, item: 'Pos'):
        current = self._block
        try:
            for index in item:
                current = current.children[index]  # 获取下一层的节点
            return current
        except (AttributeError, TypeError, IndexError):  # 中途有不是Block的节点,返回None
            return None

    def next(self):
        try:
            self._pos = next(self._pos_generator)
            return self[self._pos]
        except StopIteration:
            self.end = True
            return None

    def current(self) -> Token:
        return self[self._pos]

    def __p__template(self):
        # 如果符合语法，则返回True,对于返回False的语法判断生成器,判断器会停止该生成器
        # 语法判断主逻辑,对于每一次返回,pos都会更新
        # 比如:
        yield self.current() == ...
        yield self.current().t == ...
        yield self.current().v == ...
        # 如果返回False,判断器会停止它
        # 当有一个生成器正常退出,判断器将会将它作为匹配语法输出
        # 判断器会运行直到有一个生成器正常退出,判断器将会将它作为匹配语法输出
        # 如果没有生成器正常退出,判断器将会返回None
        yield ...  # 返回AST节点

    def _p_import(self):
        """
        import_stat ::= 'import' IDENT ('.' IDENT)* ('as' IDENT)?
        """
        yield self.current() == 'import'
        yield self.current().t == 'IDENT'
        while True:
            if self.current() == 'as':
                yield True  # 处理当前标记
                yield self.current().t == 'IDENT'  # 判断下一个标记是否为正确的标识符
                yield self.current() == '\n'  # 保证语句结束
                break  # 跳出循环,使生成器正常退出
            yield self.current() == '.'
            yield self.current().t == 'IDENT'

    def _p_from_import(self):
        """
        from_import_stat ::= 'from' IDENT ('.' IDENT)* 'import' IDENT ('as' IDENT)?
        """
        yield self.current() == 'from'
        yield self.current().t == 'IDENT'
        while True:
            if self.current().t not in ['IDENT', 'PERIOD']:
                break
            yield self.current() == '.'
            yield self.current().t == 'IDENT'
        yield self.current() == 'import'
        yield self.current().t == 'IDENT'
        if self.current() == 'as':
            yield True
            yield self.current().t == 'IDENT'

    def _p_table_def(self):
        """
        tabledef_stat ::= '[' IDENT ('.' IDENT)* ']'
        """
        yield self.current() == '['
        yield self.current().t == 'IDENT'
        while True:
            if self.current().t not in ['IDENT', 'PERIOD']:
                yield self.current() == ']'
                return
            yield self.current() == '.'
            yield self.current().t == 'IDENT'

    def parse(self, block: Block):
        self._block = block
        self._pos_generator = self._pos_generator_func()
        syntax_match_funcs = [
            getattr(self, fc)
            for fc in dir(self)
            if fc.startswith('_p_')
        ]
        # print(syntax_match_funcs)
        # self.next()
        while True:
            gm = GeneratorManager([
                fc()
                for fc in syntax_match_funcs
            ], next_hook=self.next)
            res = gm.result()
            if res:
                print(res)
            if res is None and not self.end:
                raise SyntaxError(f"Cannot match any syntax.")



from .auto import *
from .base import *
from .clock import *
from .double_dict import *
from .instance_dict import *
from .multi_set import *
from .member_manager import *
from .namespace import *
from .network_packages import *
from .once_message import *
from .permission_manager import *
from .probability_counter import *
from .timed_data import *
from .triggers import *



from ._hycore.atexit import *
from ._hycore.const import *
from ._hycore.coro_plus import *
from ._hycore.decorators import *
from ._hycore.env_plus import *
from ._hycore.file import *
from ._hycore.hash import *
from ._hycore.hytime import *
from ._hycore.import_plus import *
from ._hycore.io_addons import *
from ._hycore.json_file import *
from ._hycore.neostruct import *
from ._hycore.network import *
from ._hycore.output_methods import *
from ._hycore.path import *
from ._hycore.process import *
from ._hycore.process_addons import *
from ._hycore.socket import *
from ._hycore.sys_plus import *
from ._hycore.threading_methods import *

from ._hycore.codedoc.Struct import *
from ._hycore.data_structures import *
from ._hycore.json import *
from ._hycore.type_func import *
from ._hycore.utils import *



import sys
from pathlib import Path
from argparse import ArgumentParser
from importlib import util


class Enter:
    def __init__(self):
        self.parser = ArgumentParser()
        self.parser.add_argument('path', type=str,
                                 help='path to the utility script')

        self.parser.parse_args(sys.argv[1:2], namespace=self)
        self.root = Path(__file__).parent

    def _run_utility(self, path):
        spec = util.spec_from_file_location('utility', path)
        if spec is None or spec.loader is None:
            raise ImportError(f'Cannot load module {self.path}(One attribute is None)')
        module = util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.argv = [str(path)] + sys.argv[2:]
        module.main()

    def exec(self):
        if not self.path.endswith('.py'):
            self.path += '.py'

        path = self.root / self.path
        if not path.exists():
            raise FileNotFoundError(f'Cannot find file {self.path}')

        self._run_utility(path)


def main():
    Enter().exec()


if __name__ == '__main__':
    main()




class Tree:
    pass


class Node:
    pass


class BST(Tree):
    class _Node(Node):
        def __init__(self, value=None, left=None, right=None, parent=None, cmp=lambda a, b: a < b):
            self.value = value
            self.left = left
            self.right = right
            self.parent = parent
            self.cmp = cmp

        def insert(self, value):
            # Cmp(self.value, value) left'<' , right'>'
            # 1 -> left
            # 2 -> right
            # 0 -> value
            if self.value is None:
                self.value = value
                return
            cmp_value = self.cmp(self.value, value)
            if not cmp_value:
                self.left_check()
                self.left.insert(value)
            else:
                self.right_check()
                self.right.insert(value)

        def get_parent(self):
            return self.parent

        def query(self, value):
            if self.value == value:
                return self
            if self.cmp(value, self.value):
                return self.left.query(value) if self.left else None
            return self.right.query(value) if self.right else None

        def find_min(self):
            """Find the minimum value node in the current subtree."""
            current = self
            while current.left is not None:
                current = current.left
            return current

        def clear(self):
            # 清空值以及子树
            self.value = None
            if self.right:
                self.right.clear()
            if self.left:
                self.left.clear()

        def left_check(self):
            if self.left is None:
                self.left = self.__class__(parent=self)

        def right_check(self):
            if self.right is None:
                self.right = self.__class__(parent=self)

        def delete(self, value):
            if self.cmp(value, self.value):
                if self.left and self.right:
                    # 找到右子树中最小值
                    min_node = self.right.find_min()
                    # 删除右子树中最小值
                    self.right.delete(min_node._value)
                    # 将右子树中最小值赋值给当前节点
                    self.value = min_node._value
                elif self.left:
                    # 如果左子树不为空，则将左子树赋值给当前节点
                    self.value = self.left._value
                    # 删除左子树
                    self.left.clear()
                elif self.right:
                    # 如果右子树不为空，则将右子树赋值给当前节点
                    self.value = self.right._value
                    # 删除右子树
                    self.right.clear()
                else:
                    # 如果左右子树都为空，则将当前节点的值置为None
                    self.value = None
            elif self.cmp(value, self.value):
                if self.left:
                    self.left.delete(value)
                else:
                    raise ValueError("Value not found in tree.")
            else:
                if self.right:
                    self.right.delete(value)
                else:
                    raise ValueError("Value not found in tree.")

        def __bool__(self):
            return self.value is not None

        def __iter__(self):
            # zhongxu
            if self.left:
                yield from self.left
            yield self.value
            if self.right:
                yield from self.right

    # --------------------------------------------------------------------------------
    def __init__(self):
        self._binarytree = self._Node()

    @property
    def binarytree(self):
        return self._binarytree

    def add(self, value):
        self._binarytree.insert(value)

    def delete(self, value):
        self._binarytree.delete(value)

    def clear(self):
        self._binarytree.clear()

    def find_min(self):
        return self._binarytree.find_min()

    @property
    def left(self):
        return self._binarytree.left

    @property
    def right(self):
        return self._binarytree.right

    @property
    def value(self):
        return self._binarytree.value

    def query(self, value):
        return self._binarytree.query(value)

    def __iter__(self):
        for i in self._binarytree:
            yield i

    def __bool__(self):
        return bool(self._binarytree.value is not None)


def print_tree_img(root, linewidth=80):
    """
    打印二叉树，控制每行打印的宽度。

    :param root: 二叉树的根节点。
    :param linewidth: 每行的最大字符数，包括空格。
    """
    if not root:
        return

    # 辅助函数计算节点的字符串表示
    def node_str(node):
        return f"{node._value}" if node else ""

    # 辅助函数计算节点的宽度
    def node_width(node):
        return len(node_str(node))

    # 递归打印，采用深度优先搜索，控制每行打印宽度
    def dfs_print(node, level=0, current_line_width=0):
        if not node:
            return

        # 先序遍历打印节点
        node_str_width = node_width(node)
        if current_line_width + node_str_width > linewidth:
            print()  # 换行
            current_line_width = 0  # 重置当前行宽度计数
            level = 0  # 重置层级计数，因为换行了
        spaces = " " * (2 ** level - 1)  # 根据层级计算前置空格
        print(spaces + node_str(node), end='')
        current_line_width += node_str_width + len(spaces)

        # 递归打印左子树和右子树
        dfs_print(node.left, level + 1, current_line_width)
        dfs_print(node.right, level + 1, current_line_width)

    dfs_print(root)
    print()  # 最终换行结束输出


def print_tree(root):
    if root and root._value is not None:
        print_tree(root.left)
        print(root._value, end=' ')
        print_tree(root.right)



import threading as threading
from queue import Queue


def thread(function, args=(), kwargs={}):
    t = threading.Thread(target=function, args=args, kwargs=kwargs)
    return t


def run_new_thread(function, *args, **kwargs):
    t = thread(function, args, kwargs)
    t.start()
    return t


def run_new_daemon_thread(function, *args, **kwargs):
    t = thread(function, args, kwargs)
    t.daemon = True
    t.start()
    return t


def exit_thread(thread: threading.Thread):
    """
    This function is not safe. You should use thread.join() instead.
    """
    thread._tstate_lock.release()


def run_with_timeout(func, timeout, *args, **kwargs):
    queue = Queue()

    def target():
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            res = e

        queue.put(res)

    thread = run_new_thread(target)
    try:
        thread.join(timeout)
    except RuntimeError as e:
        raise e

    result = queue.get()

    if isinstance(result, Exception):
        raise result
    return result


def run_in_thread(func, *args, **kwargs):
    queue = Queue()

    def target():
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            res = e

        queue.put(res)

    return queue, run_new_thread(target)


def run_in_thread_with_timeout(func, timeout, *args, **kwargs):
    def target():
        return run_with_timeout(func, timeout, *args, **kwargs)

    return run_in_thread(target)



from abc import abstractmethod
from typing import Any, Callable, Optional, final


class OffsetFunction:
    @abstractmethod
    def offset(self, length): ...

    @abstractmethod
    def n_offset(self, length): ...

    @abstractmethod
    def back(self, length): ...

    @abstractmethod
    def n_back(self, length): ...

    @abstractmethod
    def offseter(self, length): ...

    @abstractmethod
    def backer(self, length): ...

    @abstractmethod
    def isend(self): ...

    @abstractmethod
    def isstart(self): ...

    @abstractmethod
    def n_offseter(self, length): ...

    @abstractmethod
    def n_backer(self, length): ...

    @abstractmethod
    def add_start(self, length): ...

    @abstractmethod
    def less_start(self, length): ...

    @abstractmethod
    def get_add_start(self, length): ...

    @abstractmethod
    def get_less_start(self, length): ...

    @abstractmethod
    def init_start(self): ...

    @abstractmethod
    def get_start_less(self, other): ...


class IterableOffsetFunction(OffsetFunction):
    @final
    def __init__(self, fat):
        self.fat = fat

    def init_start(self):
        self.fat.start = 0

    def less_start(self, length):
        self.fat.start -= length

    def add_start(self, length):
        self.fat.start += length

    def get_add_start(self, length):
        return self.fat.start + length

    def get_less_start(self, length):
        return self.fat.start - length

    def get_start_less(self, other):
        return other - self.fat.start

    def offset(self, length):
        # print(self.fat.start, length)
        if self.fat.start + length > len(self.fat.iterable):
            raise RuntimeError(
                "Cannot offset {} lengths (only {}).".format(length, len(self.fat.iterable) - self.fat.start))
        data = self.fat.iterable[self.fat.start: self.get_add_start(length)]
        self.add_start(length)
        if isinstance(data, tuple):
            return data[0]
        return data

    def n_offset(self, length):
        data = self.fat.iterable[self.fat.start: self.get_add_start(length)]
        return data

    def back(self, length):
        if self.fat.start < length:
            raise ValueError("Cannot back {} lengths (only {}).".format(length, self.fat.start))
        data = self.fat.iterable[self.get_less_start(length): self.fat.start]
        self.less_start(length)
        return data

    def n_back(self, length):
        data = self.fat.iterable[self.get_less_start(length): self.fat.start]
        return data

    def offseter(self, length):
        return self.fat.__class__(self.offset(length), self)

    def backer(self, length):
        return self.fat(self.back(length), self)

    def isend(self):
        return self.fat.start >= len(self.fat.iterable) - 1

    def isstart(self):
        return self.fat.start <= 0

    def n_offseter(self, length):
        return self.fat(self.n_offset(length), self)

    def n_backer(self, length):
        return self.fat(self.n_back(length), self)


class Offset:
    offset_func: OffsetFunction

    offset = lambda self, length: self.offset_func.offset(length)
    n_offset = lambda self, length: self.offset_func.n_offset(length)
    back = lambda self, length: self.offset_func.back(length)
    n_back = lambda self, length: self.offset_func.n_back(length)
    add_start = lambda self, length: self.offset_func.add_start(length)
    less_start = lambda self, length: self.offset_func.less_start(length)
    offseter = lambda self, length: self.offset_func.offseter(length)
    backer = lambda self, length: self.offset_func.backer(length)
    n_offseter = lambda self, length: self.offset_func.n_offseter(length)
    n_backer = lambda self, length: self.offset_func.n_backer(length)
    isend = lambda self: self.offset_func.isend()
    isstart = lambda self: self.offset_func.isstart()

    def __init__(self, iterable, offset_func: Optional[OffsetFunction] = None):
        if offset_func is None:
            offset_func = IterableOffsetFunction(self)
        self.iterable = iterable

        self.start = None

        self.offset_func = offset_func
        self.offset_func.init_start()

    def __iter__(self):
        return self.iterable[self.start:]

    def __getitem__(self, item):
        return self.iterable[item]

    def __len__(self):
        return len(self.iterable)

    def __setitem__(self, key, value):
        self.iterable[key] = value

    def to(self, _Type):
        try:
            return _Type(self.iterable)
        except TypeError as e:
            raise e from None
        except Exception as e:
            raise e from None

    def surplus(self, _Type: Optional[Callable] = None):
        if _Type is None:
            return self.n_offset(self.offset_func.get_start_less(len(self)))
        else:
            return _Type(self.surplus())

    # >
    def __gt__(self, length) -> Any:
        """
        get_offset.get_offset(length)
        :param length: an int value of length
        :return: Iterable
        """
        return self.offset(length)

    # <
    def __lt__(self, length) -> Any:
        """
        get_offset.get_offset(length)
        :param length: an int value of length
        :return: Iterable
        """
        return self.back(length)

    # >=
    def __ge__(self, length) -> Any:
        """
        get_offset.n_offset(length)
        :param length: an int value of length
        :return: Iterable
        """
        return self.n_offset(length)

    # <=
    def __le__(self, length) -> Any:
        """
        get_offset.get_offset(length)
        :param length: an int value of length
        :return: Iterable
        """
        return self.n_back(length)

    # +
    def __add__(self, length) -> Any:
        """
        start += length
        :param length: an int value of length 
        :return: None
        """
        self.add_start(length)

    # -
    def __sub__(self, length) -> Any:
        """
        start -= length
        :param length: an int value of length
        :return: None
        """
        self.less_start(length)

    # +=
    def __iadd__(self, length) -> Any:
        """
        start += length
        :param length: an int value of length
        :return: None
        """
        self.add_start(length)
        return self

    # -=
    def __isub__(self, length) -> Any:
        """
        start -= length
        :param length: an int value of length
        :return: None
        """
        self.less_start(length)
        return self

    def __lshift__(self, other):
        return self.back(other)

    def __rshift__(self, other):
        return self.offset(other)



class DoubleDict(dict):
    def __init__(self, dct=None):
        super().__init__()
        if dct:
            for k, v in dct._instances():
                self[k] = v

        self._same_values = set()

    def get(self, k, default=None):
        if k in self._same_values:
            return k

        if k not in self:
            return default

        return self[k]

    def __getitem__(self, k):
        if k in self._same_values:
            return k
        return super().__getitem__(k)

    def __setitem__(self, key, value):
        if key == value:
            self._same_values.add(key)
            return
        super().__setitem__(key, value)
        super().__setitem__(value, key)

    def __delitem__(self, key):
        if key in self._same_values:
            self._same_values.remove(key)
            return
        super().__delitem__(self[key])
        super().__delitem__(key)



from typing import Union

from ..Lexer.token import Token


class Pos:
    def __init__(self, index_ls):
        self._index_ls = index_ls  # type: list[int]

    def move_out(self):
        self._index_ls.pop()

    def move_in(self, index=0):
        self._index_ls.append(index)

    @property
    def root(self):
        return type(self)(self._index_ls[0])

    def __getitem__(self, item):
        return self._index_ls[item]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._index_ls == other._index_ls
        elif isinstance(other, list):
            return self._index_ls == other
        return False

    def __len__(self):
        return len(self._index_ls)

    def __iter__(self):
        yield from self._index_ls


class Block:
    def __init__(self, tokens: list[Union[Token, 'Block']] = None):
        self.children = tokens if tokens is not None else []

    def empty(self):
        return len(self.children) == 0

    def addChild(self, node):
        self.children.append(node)

    def sort(self):
        i = 0
        while i < len(self.children):
            if isinstance(self.children[i], Block):
                if self.children[i].empty():
                    self.children.pop(i)
                    i -= 1
                else:
                    self.children[i].sort()
            else:
                if self.children[i].type in {'WS', 'INDENT', 'DEDENT'}:
                    self.children.pop(i)
                    i -= 1
            i += 1

    def __getitem__(self, item):
        return self.children[item]

    def __len__(self):
        return len(self.children)

    def __bool__(self):
        return True

    def __setitem__(self, key, value):
        self.children[key] = value



import time
from typing import Any

import psutil


class CProcess(psutil.Process):
    def runtime(self):
        return time.time() - self.create_time()

    def as_dict(self, attrs: list[str] |
                             tuple[str, ...] |
                             set[str] |
                             frozenset[str] | None = (),
                ad_value: Any | None = ...):
        o = psutil.Process.as_dict(self, attrs, ad_value)
        o["runtime"] = self.runtime()
        return o


class ProcessInfo:
    def __init__(self, ps: CProcess):
        self._dict = ps.as_dict()

    def name(self):
        return self._dict["name"]

    @property
    def pid(self):
        return self._dict["pid"]

    def pwd(self):
        return self._dict["pwd"]

    def exe(self):
        return self._dict["exe"]

    def create_time(self):
        return self._dict["create_time"]

    def runtime(self):
        return self._dict["runtime"]

    def status(self):
        return self._dict["status"]

    def environ(self):
        return self._dict["environ"]

    def cmdline(self):
        return self._dict["cmdline"]

    def username(self):
        return self._dict["username"]

    def open_files(self):
        return self._dict["open_files"]

    def threads(self):
        return self._dict["threads"]

    @property
    def ppid(self):
        return self._dict["ppid"]

    def num_threads(self):
        return self._dict["num_threads"]

    def __str__(self):
        return f"name:{self.name()}\n" \
               f"pid:{self.pid}\n" \
               f"pwd:{self.pwd()}\n" \
               f"exe:{self.exe()}\n" \
               f"create_time:{self.create_time()}\n" \
               f"runtime:{self.runtime()}\n" \
               f"status:{self.status()}\n"

    def __repr__(self):
        return f"name:{self.name()}\n" \
               f"pid:{self.pid}\n" \
               f"pwd:{self.pwd()}\n" \
               f"exe:{self.exe()}\n" \
               f"create_time:{self.create_time()}\n" \
               f"runtime:{self.runtime()}\n" \
               f"status:{self.status()}\n"


def sys_process() -> tuple[CProcess, ...]:
    """
    return Process()
    """
    p = Any
    objects = ()
    for p in psutil.process_iter():
        try:
            objects += (CProcess(p.pid),)
        except psutil.NoSuchProcess:
            continue

    return objects


def find_processes(name):
    p = sys_process()
    objs = []
    for i in p:
        if i.name() == name:
            objs.append(i)
    return objs


def kill_process_by_name(process_name):
    ps_list = find_processes(process_name)
    for i in ps_list:
        try:
            i.terminate()
            i.wait()
        except psutil.NoSuchProcess:
            continue


def kill_process_by_pid(pid: int):
    if pid in psutil.pids():
        try:
            ps = psutil.Process(pid)
            ps.terminate()
            ps.wait()
        except psutil.NoSuchProcess:
            return False


def to_info(ps: tuple[CProcess, ...]):
    objs = []
    for i in ps:
        o = ProcessInfo(i)
        objs.append(o)
    return objs



from ..decorators import singleton_decorator


class Char(int):

    def __str__(self):
        return chr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"


@singleton_decorator
class _Null:
    ...


@singleton_decorator
class _ObjectFunc:
    def setattr(self, __self, __key, __value):
        object.__setattr__(__self, __key, __value)

    def getattr(self, __self, __value):
        return object.__getattribute__(__self, __value)

    def delattr(self, __self, __value):
        object.__delattr__(__self, __value)

    def delself(self, __self):
        ds = []
        for i in globals():
            if i is __self:
                ds.append(i)
        for i in ds:
            globals().pop(i)


null = _Null()
ObjectFunc = _ObjectFunc()
INF = float('inf')



import logging
import sys
from decimal import Decimal
from typing import Callable

outputplus_logger = logging.getLogger("HydrogenLib.OutputPlus")


def double(num, context=...):
    return Decimal(str(num), context=context)


def mapping(stream, in_min, in_max, out_min, out_max, rt_function: Callable = float):
    stream = double(stream)
    in_min, in_max, out_min, out_max = map(double, (in_min, in_max, out_min, out_max))
    outputplus_logger.debug(f"Mapping data {stream} from {in_min} {in_max} to {out_min} {out_max}")
    return rt_function(out_min + (in_max - in_min) * (stream - in_min) / (out_max - out_min))


def get_foreground(r, g, b):
    if not any((r, g, b)):
        return ''  # 不需要定义颜色
    return f"\033[38;2;{r};{g};{b}m"


def get_background(r, g, b):
    if not any((r, g, b)):
        return ''  # 不需要定义颜色
    return f"\033[48;2;{r};{g};{b}m"


def get_color_head(fr_r=0, fr_g=0, fr_b=0, bk_r=0, bk_g=0, bk_b=0):
    return get_foreground(fr_r, fr_g, fr_b) + get_background(bk_r, bk_g, bk_b)


def color_init():
    return "\033[0m"


def print_color(
        *values, sep=' ', end='\n',
        file=None, flush=False,
        foreground=None, background=None
):
    """
    :param values:
    :param sep:
    :param end:
    :param file:
    :param flush:
    :param foreground: 这个参数的实参应为一个iterable，所有值为int或其子类
    :param background: 这个参数的实参应为一个iterable，所有值为int或其子类
    :return: None
    """
    if not background:
        background = (0, 0, 0)
    if not foreground:
        foreground = (0, 0, 0)

    foreground = tuple(map(int, foreground))
    background = tuple(map(int, background))

    foreground += (0, 0, 0)

    background += (0, 0, 0)

    r, g, b = foreground[:3:]
    print(get_foreground(r, g, b), end='', file=file, flush=flush)
    r, g, b = background[:3:]
    print(get_background(r, g, b), end='', file=file, flush=flush)
    print(*values, sep=sep, end=end, file=file, flush=flush)
    print(color_init(), end='', file=file, flush=flush)


class Cursor:
    def left(self, n):
        print(f"\033[{n}D", end='', flush=True)
        return self

    def right(self, n):
        print(f"\033[{n}C", end='', flush=True)
        return self

    def up(self, n):
        print(f"\033[{n}A", end='', flush=True)
        return self

    def down(self, n):
        print(f"\033[{n}B", end='', flush=True)
        return self

    def next_line(self):
        print(f"\033[E", end='', flush=True)
        return self

    def previous_line(self):
        print(f"\033[F", end='', flush=True)
        return self

    def clear_line(self):
        print(f"\033[2K", end='', flush=True)
        return self

    def clear_screen(self):
        print(f"\033[2J", end='', flush=True)
        return self

    def clear_to_end(self):
        print(f"\033[0K", end='', flush=True)
        return self

    def clear_to_start(self):
        print(f"\033[1K", end='', flush=True)
        return self

    def move(self, x, y):
        print(f"\033[{y};{x}H", end='', flush=True)
        return self


class RedirectOutput:
    """
    重定向输出或输入至/来自IO对象或文件标识符
    例如：
    ```
        stdout, stderr = io.StringIO(), io.StringIO()
        with RedirectOutput(stdout, stderr):
            # 执行你想要执行的语句
            ...
        print(stdout.getValue(), stderr.getValue())
    ```
    """

    def __init__(self, stdout=..., stderr=..., stdin=...):
        if stdout is ...:
            stdout = sys.stdout
        if stderr is ...:
            stderr = sys.stderr
        if stdin is ...:
            stdin = sys.stdin
        self.stdout = stdout
        self.stderr = stderr
        self.stdin = stdin

        self.back = sys.stdout, sys.stderr, sys.stdin

    def redirect(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        sys.stdin = self.stdin

    def unredirect(self):
        sys.stdout, sys.stderr, sys.stdin = self.back

    def __enter__(self):
        self.redirect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unredirect()



from typing import Union, Iterable, Tuple


def mapping(stream, in_min, in_max, out_min, out_max):
    return (stream - in_min) * (out_max - out_min) / (in_max - in_min) + out_min


def pack_variable_length_int(x: int):
    """
    将整数打包为可变长格式
    """
    res = bytearray()
    while True:
        byte = x & 0x7F
        x >>= 7
        if x:
            byte |= 0x80
        res.append(byte)
        if not x:
            break
    return bytes(res)


def unpack_variable_length_int(data: Union[bytes, bytearray, Iterable[int]]) -> Tuple[int, int]:
    """
    将可变长格式的整数字节串解包
    返回: 解包后的整数, 解包使用的字节数
    """
    result = 0
    shift = 0
    count = 0
    for byte in data:
        result |= (byte & 0x7F) << shift
        shift += 7
        if byte & 0x80 == 0:
            break
        count += 1
    return result, count + 1



from ..utils.probability_counter import ProbabilityCounter


def get_probabilities(data):
    """
    Returns a dictionary of probabilities for each character in the data
    """
    pc = ProbabilityCounter()
    for i in data:
        pc.increment(i, 1)
    return pc.probabilities()


def get_probabilities_dict(data):
    """
    Returns a dictionary of probabilities for each character in the data
    """
    pc = ProbabilityCounter()
    for i in data:
        pc.increment(i, 1)
    return pc.proabilities_dict()


class HuffmanNode:
    def __init__(self):
        self.left = None
        self.right = None
        self.value = None
        self.probability = None

    def is_leaf(self):
        return self.left is None and self.right is None

    # >
    def __gt__(self, other):
        return self.probability > other.probability

    def __lt__(self, other):
        return self.probability < other.probability

    # ==
    def __eq__(self, other):
        return self.probability == other.probability

    def __str__(self):
        return f"{self.value}"

    __repr__ = __str__


class HuffmanTree:
    def __init__(self):
        self.root = None

    def build_tree(self, probabilities: dict):
        priority_queue: list[HuffmanNode] = []
        for char, probability in probabilities.items():
            node = HuffmanNode()
            node.value = char
            node.probability = probability
            priority_queue.append(node)

        while len(priority_queue) > 1:
            priority_queue.sort()
            node1 = priority_queue.pop(0)
            node2 = priority_queue.pop(0)

            prob1 = node1.probability
            prob2 = node2.probability

            new_node = HuffmanNode()
            new_node.left = node1
            new_node.right = node2
            new_node.probability = prob1 + prob2
            priority_queue.append(new_node)

        root = priority_queue.pop()
        self.root = root

    def _walk(self, node, path):
        if node.is_leaf():
            yield path, node.value
        else:
            yield from self._walk(node.left, path + '0')
            yield from self._walk(node.right, path + '1')

    def walk(self):
        return self._walk(self.root, '')


def get_huffman_codes(huffman_tree):
    return dict(huffman_tree.walk())


def get_huffman_codes_dict(huffman_tree):
    return dict(map(lambda x: (x[1], x[0]), huffman_tree.walk()))


def compress(data):
    probabilities = get_probabilities_dict(data)
    huffman_tree = HuffmanTree()
    huffman_tree.build_tree(probabilities)
    codes = get_huffman_codes_dict(huffman_tree)

    res = ''
    for char in data:
        res += codes[char]

    return res


def decompress(data, huffman_tree: HuffmanTree):
    res = ''
    current_code = huffman_tree.root
    for char in data:
        current_code = current_code.left if char == '0' else current_code.right
        if current_code.is_leaf():
            res += current_code.value
            current_code = huffman_tree.root

    return res


if __name__ == '__main__':
    test_str = "aaaaabcdef"
    huffman_tree = HuffmanTree()
    probabilities = get_probabilities_dict(test_str)
    huffman_tree.build_tree(probabilities)

    print("Test string:", repr(test_str))
    print(probabilities)

    print(*get_huffman_codes(huffman_tree).items(), sep='\n')

    compression = compress(test_str)
    print(compression)

    decompression = decompress(compression, huffman_tree)
    print(decompression)



import builtins
import typing
from collections import deque
from threading import Lock
from types import NoneType
from typing import Sequence, Optional, Union

from .methods import get_class, get_attr_bitmap_length, length_to_bytes, connect_length, get_part
from .. import abc
from ...._hycore import type_func, neostruct
from ...._hycore.type_func import get_qualname, get_subclasses_recursion


class GeneraterError(Exception):
    ...


class UnpackError(Exception):
    ...


SimpleTypes = typing.Union[int, str, bytes, float]


def pack_attr(attr_value):
    if isinstance(attr_value, SimpleTypes):
        data = neostruct.neopack(attr_value)
        type_name = type_func.get_type_name(attr_value)
    elif isinstance(attr_value, BinStructBase):
        type_name = get_qualname(attr_value)
        data = attr_value.pack()
    elif isinstance(attr_value, Sequence):
        type_name = type_func.get_type_name(attr_value)
        if type_name not in ['list', 'tuple', 'set']:
            raise NotImplementedError(f'Unsupported type: {type(type_name)}')
        data = _pack_sequence(attr_value)
    elif isinstance(attr_value, dict):
        type_name = 'dict'
        data = _pack_dict(attr_value)
    elif isinstance(attr_value, NoneType):
        type_name = 'NoneType'
        data = None
    else:
        raise NotImplementedError(f'Unsupported type: {type(attr_value)}')

    return connect_length(type_name.encode()) + connect_length(data, no_none=False)


def unpack_attr(offset):
    type_name = get_part(offset).decode()
    if type_name == 'NoneType':
        return None
    packed_data = get_part(offset)
    if type_name in bin_types:
        origin_data = BinStructBase.unpack(packed_data)
    elif hasattr(builtins, type_name):
        type_ = getattr(builtins, type_name)
        if issubclass(type_, SimpleTypes):
            origin_data = neostruct.neounpack(type_, packed_data)
        elif issubclass(type_, list):
            origin_data = _unpack_sequence(packed_data)
        elif issubclass(type_, dict):
            origin_data = _unpack_dict(packed_data)
        else:
            raise NotImplementedError(f'Unsupported built-in type: {type_name}')
    else:
        raise NotImplementedError(f'Unsupported type: {type_name}')

    return origin_data


def _pack_sequence(seq):
    sequence_bytes = b''
    for item in seq:
        sequence_bytes += pack_attr(item)

    return sequence_bytes


def _unpack_sequence(data):
    offset = type_func.Offset(data)
    ls = deque()
    while not offset.isend():
        ls.append(unpack_attr(offset))

    return list(ls)


def _pack_dict(dic: dict):
    dict_bytes = b''
    for key, value in dic.items():
        dict_bytes += pack_attr(key)
        dict_bytes += pack_attr(value)
    return dict_bytes


def _unpack_dict(data):
    offset = type_func.IndexOffset.Offset(data)
    result_dict = {}
    while not offset.isend():
        key = unpack_attr(offset)
        value = unpack_attr(offset)
        result_dict[key] = value
    return result_dict


class BinStructBase:
    """

    # 二进制结构体基类

    方法:
        - pack() **不可覆写**, 打包结构体
        - unpack(data) 解包结构体
    属性:
        - _data_ 需要打包的属性**列表**

    """
    __data__ = []  # Variables' names

    def pack_event(self, *args, **kwargs):
        """
        打包事件,进行打包前的处理
        """
        return True

    def pack_attr_event(self, attr_name: str):
        return getattr(self, attr_name)

    def unpack_event(self, *args, **kwargs) -> Optional[Union[Exception, bool]]:
        """
        解包事件,解包后对原始数据的重新处理
        """
        return True

    def __init__(self, *args, **kwargs):
        names = set()
        for name, value in zip(self.__data__, args):
            setattr(self, name, value)
            names.add(name)

        for name, value in kwargs.items():
            if name in names:
                raise ValueError(f'Duplicate name: {name}')
            setattr(self, name, value)

        self.serializer_funcs = {
            'pack': pack_attr,
            'unpack': unpack_attr
        }

    # @final
    def pack(self, *args, **kwargs):
        """
        打包结构体
        :param args: 构造函数参数
        :param kwargs: 构造函数参数
        """

        # | Name | AttrMapping | (Value, Type) pairs |

        pack_event = self.pack_event(*args, **kwargs)
        if pack_event is not True:
            raise GeneraterError('Pack event failed', pack_event)

        __data__ = self.__data__

        pack_attr, unpack_attr = self.serializer_funcs.get('pack'), self.serializer_funcs.get('unpack')
        if not (pack_attr and unpack_attr):
            raise NotImplementedError('Pack attr or unpack attr is not defined')

        this_name = get_qualname(self)
        this_length_head = length_to_bytes(this_name)

        part_name = this_length_head + this_name.encode()

        bitmap = type_func.Bitmap()

        for index, attr in enumerate(__data__):
            origin_data = getattr(self, attr)
            if hasattr(self.__class__, attr):
                if origin_data is getattr(self.__class__, attr):
                    continue  # 未修改的属性, 跳过
            bitmap[index] = True

        part_bitmap = bitmap.pack()

        part_kvpairs = b''

        for i, attr in enumerate(__data__):
            on = bitmap[i]
            if not on:
                continue

            value = self.pack_attr_event(attr)
            current_part = pack_attr(value)
            part_kvpairs += current_part

        packed_data = (
                part_name + part_bitmap + part_kvpairs
        )
        # print(
        #     'Pack:::\n\t',
        #     'Part::Name=', part_name,
        #     '\n\t',
        #     'Part::Bitmap=', part_bitmap, bitmap,
        #     '\n\t',
        #     'Part::KVPairs=', part_kvpairs
        # )

        return packed_data

    @staticmethod
    # @final
    def unpack(data, __data__=None, *args, **kwargs):
        """
        解包结构体.
        :param data: 原始数据
        :param __data__: 构造函数参数
        :param args: 传递给unpack_event函数的参数
        :param kwargs: 传递给unpack_event函数的参数
        """

        offset = type_func.IndexOffset.Offset(data)
        this_name = get_part(offset).decode()

        # print(this_name)
        typ = get_class(this_name)
        # print(typ)

        __data__ = __data__ or typ.__data__

        bitmap_length = get_attr_bitmap_length(len(__data__))
        bitmap = type_func.Bitmap.unpack(offset >> bitmap_length)

        # 获取已设置的属性
        set_attrs = [attr_name for attr_name, on in zip(__data__, bitmap) if on]
        # print(set_attrs)
        temp_dct = {}

        attr_count = 0
        while not offset.isend():
            origin_data = unpack_attr(offset)
            temp_dct[attr_count] = origin_data
            attr_count += 1

        attr_dct = {}

        for attr, value in temp_dct.items():
            attr_dct[set_attrs[attr]] = value

        ins = typ(**attr_dct)  # type: BinStructBase
        Res = ins.unpack_event(*args, **kwargs)
        if Res is not None:
            if type_func.is_errortype(Res):
                raise UnpackError('An error occurred during unpacking') from Res
        return ins

    def mini_pack(self, data):
        return self.serializer_funcs['pack'](data)

    def mini_unpack(self, data):
        offset = type_func.IndexOffset.Offset(data)
        return self.serializer_funcs['unpack'](offset)  # Unpack 的参数是一个索引偏移对象

    @classmethod
    def to_struct(cls, obj, __data__=None):
        """
        根据传入的对象以及__data__列表,构建结构体
        """
        if isinstance(obj, BinStructBase):
            if __data__ is None:
                return obj
            elif __data__ == obj.__data__:
                return obj
            else:
                raise GeneraterError('无法确定结构体需要包含的属性')
        if __data__ is None:
            if hasattr(obj, '__data__'):
                __data__ = getattr(obj, '__data__')

        if __data__ is None:
            raise GeneraterError('无法确定结构体需要包含的属性')

        ins = cls(**{name: getattr(obj, name) for name in __data__})
        ins.__data__ = __data__
        return ins

    @classmethod
    def is_registered(cls):
        """
        检查此类是否已经注册
        """
        return get_qualname(cls) in bin_types

    @property
    def __attrs_dict(self):
        dct = {}
        for name in self.__data__:
            dct[name] = getattr(self, name)
        return dct

    def __str__(self):
        kv_pairs = list(self.__attrs_dict.items())
        return f'{self.__class__.__name__}({", ".join((f"{name}={repr(value)}" for name, value in kv_pairs))})'

    def __eq__(self, other):
        if not isinstance(other, BinStructBase):
            return False
        return self.__attrs_dict == other.__attrs_dict

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(tuple(self.__attrs_dict.items()))

    __repr__ = __str__


class Struct(abc.Serializer):
    struct = BinStructBase

    def dumps(self, data: Union[BinStructBase, typing.Any]):
        if isinstance(data, self.struct):
            return data.pack()
        else:
            return self.struct.mini_pack(data)

    def loads(self, data, __data__=None, mini=False):
        if mini is True:
            return self.struct.mini_unpack(data)
        elif mini is False:
            return self.struct.unpack(data, __data__=__data__)
        else:
            try:
                return self.struct.unpack(data, __data__=__data__)
            except:
                return self.struct.mini_unpack(data)


bin_types = {
    get_qualname(BinStructBase),
}

_flush_lock = Lock()


def flush_bin_types():
    """
    为了保证自定义的BinStruct子类能被正确解析,需要调用此函数刷新结构体注册表
    """
    with _flush_lock:
        global bin_types
        bin_types |= set((get_qualname(cls) for cls in get_subclasses_recursion(BinStructBase)))


def get_bin_types():
    return bin_types.copy()



from typing import Callable
from threading import Lock


class SignalInstance:
    def __init__(self):
        self.hooks = []
        self._lck = Lock()

    def connect(self, func):
        # with self._lck:
            self.hooks.append(func)

    def disconnect(self, func):
        # with self._lck:
            self.hooks.remove(func)

    def clear(self):
        # with self._lck:
            self.hooks.clear()

    def emit(self, *args, **kwargs):
        with self._lck:
            for hook in self.hooks.copy():  # Copy保证线程安全
                hook(*args, **kwargs)


class Signal:
    _instance_to_signal = {}

    def __get_signal(self, instance):
        key = id(instance)
        if key not in self._instance_to_signal:
            signal = SignalInstance()
            self._instance_to_signal[key] = signal
        return self._instance_to_signal[key]

    def __get__(self, instance, owner):
        if isinstance is None:
            return self
        return self.__get_signal(instance)

    def __delete__(self, instance):
        self._instance_to_signal.clear()

    def __init__(self):
        self._instance_to_signal = {}


class Hook:
    def __init__(self, func):
        self._pre = []
        self._fc = func
        self._post = []

    def pre(self, func):
        self._pre.append(func)

    def post(self, func):
        self._post.append(func)

    def get(self, name) -> list[Callable]:
        if name == 'pre':
            return self._pre
        elif name == 'post':
            return self._post
        else:
            raise ValueError('name must be "pre" or "post"')

    def __call__(self, *args, **kwargs):
        for func in self._pre:
            func(*args, **kwargs)
        res = self._fc(*args, **kwargs)
        for func in self._post:
            func(res, *args, **kwargs)  # 添加返回值




import re

import win32file
import win32pipe


def is_pipe_path(path: str):
    r"""
    检查是否是一个标准的pipe路径

    一个标准的Pipe路径应为:``\\.\pipe\pipe_name``

    或者：``\\?\pipe\pipe_name``

    且，pipe路径不允许包含子文件夹，Pipe路径中的pipe_name不能为空


    :param path: 待检查的Pipe路径
    :return: Bool值，表示是否为标准Pipe路径
    """
    match_str = re.match(r"(\\\\([.?])\\pipe\\)[^\\/\s()\"\[\]*$%^`]+$", path)
    return match_str is not None


def format_pipe_name(name: str):
    """
    格式化Pipe名称为Pipe路径
    如果传入参数被确认已经是Pipe路径，那么直接返回
    :param name: 待格式化的名称，不能为空字符序列
    :return:
    """
    if name.isspace():
        raise ValueError("Pipe name cannot be empty")
    if is_pipe_path(name):
        return name
    return fr"\\.\pipe\{name}"


def get_pipe_object(name, buffer_size=65535, max_instances=1):
    return win32pipe.CreateNamedPipe(
        format_pipe_name(name),
        win32pipe.PIPE_ACCESS_DUPLEX,
        win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE | win32pipe.PIPE_WAIT,
        max_instances,
        buffer_size,
        buffer_size,
        0,
        None
    )


def get_pipe_handle(name):
    return win32file.CreateFile(
        format_pipe_name(name),
        win32file.GENERIC_READ | win32file.GENERIC_WRITE,
        0,
        None,
        win32file.OPEN_EXISTING,
        0,
        None
    )


class _Reader:
    def __init__(self, name):
        self.name = name
        self.filename = format_pipe_name(name)


if __name__ == '__main__':
    print(is_pipe_path(r"\\.\pipe\Pipe"))
    print(is_pipe_path(r"\\.\pipe\Pipe\adadf"))
    print(is_pipe_path(r"\\.\pipe\x df"))
    print(is_pipe_path(r"\\.\pipe\&dfaf"))



class Interpreter:
    def __init__(self, ast):
        self.ast = ast
        self.environment = {}
        self.tables = {}

    def evaluate(self, node):
        if node.type == 'Literal':
            return node._env
        elif node.type == 'Identifier':
            return self.environment[node._env]
        elif node.type == 'Expression':
            try:
                value = eval(node._env, self.environment)
                return value
            except Exception as e:
                raise RuntimeError(f"Error evaluating expression {node._env}: {e}")
        elif node.type == 'Table':
            if node._env not in self.tables:
                self.tables[node._env] = {}
            if node.children:
                for child in node.children:
                    if child.type == 'Pass':
                        continue
                    self.evaluate(child)
            return self.tables[node._env]
        elif node.type == 'Assignment':
            left = self.evaluate(node.children[0])
            right = self.evaluate(node.children[1])
            if isinstance(left, dict):
                key = node.children[0]._env
                self.tables[key][right] = right
            else:
                self.environment[left] = right
        elif node.type == 'ImportStatement':
            module_name = node._env
            module = __import__(module_name)
            self.environment[module_name] = module
        elif node.type == 'FromImportStatement':
            module_name, name, alias = node._env
            module = __import__(module_name)
            if alias:
                self.environment[alias] = getattr(module, name)
            else:
                self.environment[name] = getattr(module, name)
        elif node.type == 'ExpressionStatement':
            return self.evaluate(node.children[0])
        elif node.type == 'Block':
            for statement in node.children:
                self.evaluate(statement)
        elif node.type == 'Pass':
            pass
        else:
            raise ValueError(f"Unknown node type: {node.type}")

    def run(self):
        self.evaluate(self.ast)


    def get_result(self):
        return self.tables, self.environment



from abc import ABC, abstractmethod


class AbstractDynamicModel(ABC):
    callback_metaclass: object
    running: bool

    @abstractmethod
    def update(self) -> None:
        """
        更新模型
        """
        ...

    def init(self):
        """
        初始化模型
        """
        self.running = True  # 默认初始化功能

    def stop(self):
        """
        停止并释放模型
        """
        self.running = False  # 默认释放功能


class AbstractExecutor(ABC):
    @abstractmethod
    def on_run(self):
        """
        当模型与执行器绑定并被启动时,调用此方法
        """
        ...

    @abstractmethod
    def on_stop(self):
        """
        当模型返回结束消息时,调用此方法
        """
        ...



# Generated from antlr4_grammar/hyconfiglanguage.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .hyconfiglanguageParser import hyconfiglanguageParser
else:
    from hyconfiglanguageParser import hyconfiglanguageParser

# This class defines a complete listener for a parse tree produced by hyconfiglanguageParser.
class hyconfiglanguageListener(ParseTreeListener):

    # Enter a parse tree produced by hyconfiglanguageParser#configFile.
    def enterConfigFile(self, ctx:hyconfiglanguageParser.ConfigFileContext):
        pass

    # Exit a parse tree produced by hyconfiglanguageParser#configFile.
    def exitConfigFile(self, ctx:hyconfiglanguageParser.ConfigFileContext):
        pass


    # Enter a parse tree produced by hyconfiglanguageParser#definition.
    def enterDefinition(self, ctx:hyconfiglanguageParser.DefinitionContext):
        pass

    # Exit a parse tree produced by hyconfiglanguageParser#definition.
    def exitDefinition(self, ctx:hyconfiglanguageParser.DefinitionContext):
        pass


    # Enter a parse tree produced by hyconfiglanguageParser#tableDef.
    def enterTableDef(self, ctx:hyconfiglanguageParser.TableDefContext):
        pass

    # Exit a parse tree produced by hyconfiglanguageParser#tableDef.
    def exitTableDef(self, ctx:hyconfiglanguageParser.TableDefContext):
        pass


    # Enter a parse tree produced by hyconfiglanguageParser#varDef.
    def enterVarDef(self, ctx:hyconfiglanguageParser.VarDefContext):
        pass

    # Exit a parse tree produced by hyconfiglanguageParser#varDef.
    def exitVarDef(self, ctx:hyconfiglanguageParser.VarDefContext):
        pass


    # Enter a parse tree produced by hyconfiglanguageParser#importStmt.
    def enterImportStmt(self, ctx:hyconfiglanguageParser.ImportStmtContext):
        pass

    # Exit a parse tree produced by hyconfiglanguageParser#importStmt.
    def exitImportStmt(self, ctx:hyconfiglanguageParser.ImportStmtContext):
        pass



del hyconfiglanguageParser


from ._types import Block
from ...._hycore.data_structures import Stack
from ...._hycore.type_func import ConstDict

bracket_map = ConstDict(
    {
        '(': ')',
        '[': ']',
        '{': '}',
    }
)


class BracketTracker:
    def __init__(self):
        self._s = Stack()

    def trace(self, char):
        if char in bracket_map.keys():
            self._s.push(char)
        elif char in bracket_map.values():
            top = self._s.at_top
            if bracket_map[top] != char:
                raise SyntaxError(f"Bracket mismatch. (Except '{bracket_map[top]}', but got '{char}')")
            self._s.pop()

    def stack_length(self):
        return self._s.size()

    def matched(self):
        return self._s.is_empty()


class BlockParser:
    def __init__(self):
        self.tokens = None

    def __parse_to_blocks(self):
        stack = Stack([Block()])
        tracker = BracketTracker()
        for token in self.tokens:
            if token.type in ['LP', 'RP']:
                tracker.trace(token.value)
            if token.type == 'INDENT' and tracker.matched():
                new_block = Block()
                if stack.at_top:
                    stack.at_top.addChild(new_block)
                stack.push(new_block)
            elif token.type == 'DEDENT' and tracker.matched():
                stack.pop()
            else:
                if stack.at_top:
                    stack.at_top.addChild(token)
                else:
                    stack.push(Block([token]))
        return stack.at_top

    def parse(self, tokens):
        self.tokens = tokens
        return self.__parse_to_blocks()



# Generated from antlr4_grammar/hyconfiglanguage.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,8,39,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,1,0,5,0,12,8,0,
        10,0,12,0,15,9,0,1,0,1,0,1,1,1,1,1,1,3,1,22,8,1,1,2,1,2,1,2,1,2,
        1,3,1,3,1,3,1,3,1,4,1,4,3,4,34,8,4,1,4,1,4,1,4,1,4,0,0,5,0,2,4,6,
        8,0,0,37,0,13,1,0,0,0,2,21,1,0,0,0,4,23,1,0,0,0,6,27,1,0,0,0,8,33,
        1,0,0,0,10,12,3,2,1,0,11,10,1,0,0,0,12,15,1,0,0,0,13,11,1,0,0,0,
        13,14,1,0,0,0,14,16,1,0,0,0,15,13,1,0,0,0,16,17,5,0,0,1,17,1,1,0,
        0,0,18,22,3,4,2,0,19,22,3,6,3,0,20,22,3,8,4,0,21,18,1,0,0,0,21,19,
        1,0,0,0,21,20,1,0,0,0,22,3,1,0,0,0,23,24,5,1,0,0,24,25,5,6,0,0,25,
        26,5,2,0,0,26,5,1,0,0,0,27,28,5,6,0,0,28,29,5,3,0,0,29,30,5,7,0,
        0,30,7,1,0,0,0,31,32,5,4,0,0,32,34,5,6,0,0,33,31,1,0,0,0,33,34,1,
        0,0,0,34,35,1,0,0,0,35,36,5,5,0,0,36,37,5,6,0,0,37,9,1,0,0,0,3,13,
        21,33
    ]

class hyconfiglanguageParser ( Parser ):

    grammarFileName = "hyconfiglanguage.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'['", "']'", "'='", "'from'", "'import'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "NAME", "VALUE", "INDENT" ]

    RULE_configFile = 0
    RULE_definition = 1
    RULE_tableDef = 2
    RULE_varDef = 3
    RULE_importStmt = 4

    ruleNames =  [ "configFile", "definition", "tableDef", "varDef", "importStmt" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    NAME=6
    VALUE=7
    INDENT=8

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ConfigFileContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(hyconfiglanguageParser.EOF, 0)

        def definition(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(hyconfiglanguageParser.DefinitionContext)
            else:
                return self.getTypedRuleContext(hyconfiglanguageParser.DefinitionContext,i)


        def getRuleIndex(self):
            return hyconfiglanguageParser.RULE_configFile

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterConfigFile" ):
                listener.enterConfigFile(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitConfigFile" ):
                listener.exitConfigFile(self)




    def configFile(self):

        localctx = hyconfiglanguageParser.ConfigFileContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_configFile)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 13
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 114) != 0):
                self.state = 10
                self.definition()
                self.state = 15
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 16
            self.match(hyconfiglanguageParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DefinitionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def tableDef(self):
            return self.getTypedRuleContext(hyconfiglanguageParser.TableDefContext,0)


        def varDef(self):
            return self.getTypedRuleContext(hyconfiglanguageParser.VarDefContext,0)


        def importStmt(self):
            return self.getTypedRuleContext(hyconfiglanguageParser.ImportStmtContext,0)


        def getRuleIndex(self):
            return hyconfiglanguageParser.RULE_definition

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDefinition" ):
                listener.enterDefinition(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDefinition" ):
                listener.exitDefinition(self)




    def definition(self):

        localctx = hyconfiglanguageParser.DefinitionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_definition)
        try:
            self.state = 21
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 18
                self.tableDef()
                pass
            elif token in [6]:
                self.enterOuterAlt(localctx, 2)
                self.state = 19
                self.varDef()
                pass
            elif token in [4, 5]:
                self.enterOuterAlt(localctx, 3)
                self.state = 20
                self.importStmt()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TableDefContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(hyconfiglanguageParser.NAME, 0)

        def getRuleIndex(self):
            return hyconfiglanguageParser.RULE_tableDef

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTableDef" ):
                listener.enterTableDef(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTableDef" ):
                listener.exitTableDef(self)




    def tableDef(self):

        localctx = hyconfiglanguageParser.TableDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_tableDef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 23
            self.match(hyconfiglanguageParser.T__0)
            self.state = 24
            self.match(hyconfiglanguageParser.NAME)
            self.state = 25
            self.match(hyconfiglanguageParser.T__1)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VarDefContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(hyconfiglanguageParser.NAME, 0)

        def VALUE(self):
            return self.getToken(hyconfiglanguageParser.VALUE, 0)

        def getRuleIndex(self):
            return hyconfiglanguageParser.RULE_varDef

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVarDef" ):
                listener.enterVarDef(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVarDef" ):
                listener.exitVarDef(self)




    def varDef(self):

        localctx = hyconfiglanguageParser.VarDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_varDef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 27
            self.match(hyconfiglanguageParser.NAME)
            self.state = 28
            self.match(hyconfiglanguageParser.T__2)
            self.state = 29
            self.match(hyconfiglanguageParser.VALUE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ImportStmtContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self, i:int=None):
            if i is None:
                return self.getTokens(hyconfiglanguageParser.NAME)
            else:
                return self.getToken(hyconfiglanguageParser.NAME, i)

        def getRuleIndex(self):
            return hyconfiglanguageParser.RULE_importStmt

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterImportStmt" ):
                listener.enterImportStmt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitImportStmt" ):
                listener.exitImportStmt(self)




    def importStmt(self):

        localctx = hyconfiglanguageParser.ImportStmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_importStmt)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==4:
                self.state = 31
                self.match(hyconfiglanguageParser.T__3)
                self.state = 32
                self.match(hyconfiglanguageParser.NAME)


            self.state = 35
            self.match(hyconfiglanguageParser.T__4)
            self.state = 36
            self.match(hyconfiglanguageParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import jsonpickle


def encode(data, indent=None):
    """
    return jsonpickle.encode(data, indent=indent)
    """
    return jsonpickle.encode(data, indent=indent)


def decode(data):
    """
    return jsonpickle.decode(data)
    """
    return jsonpickle.decode(data)


def dump(data, file, *args, **kwargs):
    """
    编码对象并储存到文件
    """
    with open(file, 'w') as f:
        f.write(
            jsonpickle.encode(data, *args, **kwargs)
        )


def load(file, *args, **kwargs):
    """
    从文件解码对象
    """
    return jsonpickle.decode(
        open(file).read(),
        *args, **kwargs
    )



from abc import ABC, abstractmethod
from typing import Any, final


class Serializer(ABC):
    _Instance = None

    @final
    def __new__(cls, *args, **kwargs):  # 序列器一定是单例的
        if cls._Instance is None:
            cls._Instance = super().__new__(cls, *args, **kwargs)
        return cls._Instance

    @abstractmethod
    def dumps(self, data) -> bytes:
        ...

    @abstractmethod
    def loads(self, data) -> Any:
        ...



from os import PathLike
from pathlib import Path
from typing import Literal, Optional, Union
import os


def path_to(*path):
    path = os.sep.join(path)
    return path


def isdir(path):
    return os.path.isdir(path)


def isfile(file):
    return os.path.isfile(file)


def exists(path):
    return os.path.exists(path)


def mkdir(path):
    os.mkdir(path)


def mkdirs(path):
    if isdir(path):
        return
    os.makedirs(path)


def rmdir(path):
    os.rmdir(path)


def remove(path):
    os.remove(path)


def rename(old, new):
    os.rename(old, new)


def mkfile(file, clear=True):
    if not isfile(file):
        directory = os.path.dirname(file)
        if directory:
            mkdirs(directory)
        open(file, "w").close()
    elif clear:
        open(file, "w").close()


def listdir(path):
    return os.listdir(path)


def scandir(path):
    for i in os.scandir(path):
        try:
            yield i
        except PermissionError:
            continue


def scandir_ls(path):
    return list(scandir(path))


def rmdirs(path):
    for i in scandir(path):
        if i.is_dir():
            rmdirs(i.path)
        else:
            remove(i.path)
    rmdir(path)


def rmfile(file):
    remove(file)


def abspath(path):
    return os.path.abspath(path)


def isabspath(path):
    return os.path.isabs(path)


def split(path):
    if isinstance(path, (list, tuple)):
        return list(path)
    else:
        path = process(path)
        pls = list(os.path.split(path))
        # 去重 \\
        return pls


def process(path) -> str:
    path = str(path)
    path = path.replace("/", "\\")
    path = path.replace("\\\\", "\\")
    return path


def _tree(folder) -> dict:
    f_dic = {folder: {}}
    now_dic = f_dic[folder]
    folder = path_to(folder)

    if isfile(folder):
        now_dic[folder] = os.path.getsize(folder)
    else:
        for i in scandir(folder):
            if i.is_dir():
                now_dic[i.path] = tree(i.path)
            else:
                now_dic[i.path] = os.path.getsize(i.path)
    return f_dic


def tree(folder):
    t = _tree(folder)
    return t




class StringPath(str):
    def __init__(self, __dir=None):
        if __dir is None:
            self._path = None
        else:
            self._path = Path(__dir)

    def get_path(self) -> Path:
        return self._path

    def using(self, path):
        return self._path / path

    def absolute(self):
        return self.__class__(self._path.absolute())

    def as_uri(self):
        return self.__class__(self._path.as_uri())

    def as_posix(self):
        return self.__class__(self._path.as_posix())

    @property
    def anchor(self):
        return self.__class__(self._path.anchor)

    def chmod(self, mode, *, follow_symlinks=True):
        self._path.chmod(mode, follow_symlinks=follow_symlinks)

    @classmethod
    def cwd(cls):
        return cls(Path.cwd())

    @property
    def drive(self):
        return self._path.drive

    def exists(self, follow_symlinks=True):
        return self._path.exists(follow_symlinks=follow_symlinks)

    def expanduser(self):
        return self._path.expanduser()

    def glob(self, pattern, *, case_sensitive: Optional[bool] = None):
        return self._path.glob(pattern, case_sensitive=case_sensitive)

    def group(self):
        return self._path.group()

    def hardlink_to(self, target):
        self._path.hardlink_to(target)

    @classmethod
    def home(cls):
        return cls(Path.home())

    def is_absolute(self):
        return self._path.is_absolute()

    def is_block_device(self):
        return self._path.is_block_device()

    def is_char_device(self):
        return self._path.is_char_device()

    def is_dir(self):
        return self._path.is_dir()

    def is_fifo(self):
        return self._path.is_fifo()

    def is_file(self):
        return self._path.is_file()

    def is_junction(self):
        return self._path.is_junction()

    def is_mount(self):
        return self._path.is_mount()

    def is_relative_to(self, other, *_deprecated):
        return self._path.is_relative_to(other, *_deprecated)

    def is_reserved(self):
        return self._path.is_reserved()

    def is_socket(self):
        return self._path.is_socket()

    def is_symlink(self):
        return self._path.is_symlink()

    def iterdir(self):
        return self._path.iterdir()

    def joinpath(self, *args):
        return self.__class__(self._path.joinpath(*args))

    def lchmod(self, mode):
        self._path.lchmod(mode)

    def lstat(self):
        return self._path.lstat()

    def match(self, pattern, *, case_sensitive: Optional[bool] = None):
        return self._path.match(pattern, case_sensitive=case_sensitive)

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        self._path.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    @property
    def name(self):
        return self._path.name

    def open(self,
             mode: Literal[
                 "r+", "+r", "rt+", "r+t", "+rt", "tr+", "t+r", "+tr", "w+", "+w", "wt+", "w+t", "+wt", "tw+", "t+w",
                 "+tw", "a+", "+a", "at+", "a+t", "+at", "ta+", "t+a", "+ta", "x+", "+x", "xt+", "x+t", "+xt", "tx+",
                 "t+x", "+tx", "w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx", "r", "rt", "tr", "U", "rU", "Ur",
                 "rtU", "rUt", "Urt", "trU", "tUr", "Utr"
             ] = "r",
             buffering: int = -1,
             encoding: Optional[str] = None,
             errors: Optional[str] = None,
             newline: Optional[str] = None):
        return self._path.open(
            mode=mode, buffering=buffering,
            encoding=encoding, errors=errors, newline=newline,
        )

    def owner(self):
        return self._path.owner()

    @property
    def parents(self):
        return [self.__class__(pre) for pre in self._path.parents]

    @property
    def parts(self):
        return self._path.parts

    def read_bytes(self):
        return self._path.read_bytes()

    def read_text(self, encoding=None, errors=None):
        return self._path.read_text(encoding=encoding, errors=errors)

    def readlink(self):
        return self._path.readlink()

    def relative_to(self,
                    other: Union[str, PathLike[str]],
                    *_deprecated: Union[str, PathLike[str]],
                    walk_up: bool = False):
        return self.__class__(
            self._path.relative_to(
                other,
                *_deprecated,
                walk_up=walk_up
            )
        )

    def rename(self, target):
        self._path.rename(target)

    def path_replace(self, target):
        self._path.replace(target)

    def resolve(self, strict=False):
        return self.__class__(self._path.resolve(strict=strict))

    def rglob(self, pattern, *, case_sensitive: Optional[bool] = None):
        return self._path.rglob(pattern, case_sensitive=case_sensitive)

    def rmdir(self):
        self._path.rmdir()

    def root(self):
        return self._path.root

    def samefile(self, other_path):
        return self._path.samefile(other_path)

    def stat(self):
        return self._path.stat()

    @property
    def stem(self):
        return self._path.stem

    @property
    def suffix(self):
        return self._path.suffix

    @property
    def suffixes(self):
        return self._path.suffixes

    def symlink_to(self, target, target_is_directory=False):
        self._path.symlink_to(target, target_is_directory=target_is_directory)

    def touch(self, mode=0o666, exist_ok=True):
        self._path.touch(mode=mode, exist_ok=exist_ok)

    def unlink(self, missing_ok=False):
        self._path.unlink(missing_ok=missing_ok)

    def walk(self, topdown=True, onerror=None, followlinks=False):
        return self._path.walk(topdown, onerror, followlinks)

    def with_name(self, name):
        return self.__class__(self._path.with_name(name))

    def with_stem(self, stem):
        return self.__class__(self._path.with_stem(stem))

    def with_segments(self, *args):
        return self.__class__(self._path.with_segments(*args))

    def with_suffix(self, suffix: str):
        return self.__class__(self._path.with_suffix(suffix))

    def write_bytes(self, data):
        self._path.write_bytes(data)

    def write_text(self, data, encoding=None, errors=None, newline=None):
        self._path.write_text(data, encoding=encoding, errors=errors, newline=newline)

    @property
    def parent(self):
        return self.__class__(self._path.parent)

    def __truediv__(self, path):
        return self.__class__(self._path / path)

    def __str__(self):
        return str(self._path)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._path})'



from ...abc.backend import BackendABC
from ....hystruct import *
from ....hycore import json_types


class Json_Backend(BackendABC):
    serializer = Json()
    support_types = (json_types, )

    def save(self):
        with self._fd.open(self.file, 'wb') as f:
            f.write(self.serializer.dumps(self._data))

    def load(self):
        with self._fd.open(self.file, 'rb') as f:
            if f.size:
                self.existing = True
                dic = self.serializer.loads(f.read())
                self.init(**dic)



import importlib
import importlib.util
import os
import sys


def load_source(name, path):
    if name in sys.modules:
        raise ImportError(f"Module {name} already exists in sys.modules")

    if not os.path.exists(path):
        raise FileNotFoundError(f"Cannot find file {path}")

    spec = importlib.util.spec_from_file_location(name, path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot load module {name}(One attribute is None)")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_source_noname(path):
    """
    use `os.path.basename` to get module name
    :param path: source code file path
    """
    name = os.path.basename(path)
    index = name.rfind(".")
    if index == -1:
        raise ValueError(f"Invalid file name {name}")
    name = name[:index]
    return load_source(name, path)


def load_package(name, path):
    if name in sys.modules:
        raise ImportError(f"Package {name} already exists in sys.modules")

    package_path = path + os.sep + name
    if not os.path.exists(package_path):
        raise FileNotFoundError(
            f"Cannot find package {name} at {package_path}")

    init_filename = os.path.join(package_path, "__init__.py")
    if not os.path.exists(init_filename):
        raise FileNotFoundError(
            f"Cannot find __init__.py in package {name}")

    spec = importlib.util.spec_from_file_location(name, init_filename)
    if spec is None:
        raise ImportError(f"Cannot load package {name}(Spec by None)")
    if spec.loader is None:
        raise ImportError(f"Cannot load package {name}(Loader by None)")

    package = importlib.util.module_from_spec(spec)

    sys.modules[name] = package
    spec.loader.exec_module(package)

    return package


def import_source(path):
    path, name = os.path.split(path)
    return load_source(name[:-3], path)


def import_package(path):
    name = os.path.basename(path)
    path = os.path.dirname(path)
    return load_package(name, path)



import inspect
import queue
import threading
import time
from typing import Union


class Node:
    def __init__(self, leaf=True):
        self.leaf = leaf
        self.n = 0
        self.keys = []
        self.values = []
        self.children = []

    def __del__(self):
        for child in self.children:
            del child

    def __str__(self):
        if self.leaf:
            return f"Leaf Node({dict(zip(self.keys, self.values))})"
        else:
            return f"Internal Node({dict(zip(self.keys, [child.__str__() for child in self.children]))})"

    def to_dict(self):
        root = {
            'keys': self.keys,
            'values': self.values,
            'leaf': self.leaf,
            'n': self.n,
            'children': [
                i.to_dict() for i in self.children
            ]
        }

    @classmethod
    def from_dict(cls, bp_dict: dict):
        node = cls()
        node.keys, node.values = bp_dict['keys'], bp_dict['values']
        node.leaf = bp_dict['leaf']
        node.n = bp_dict['n']

        node.children = []

        for i in range(len(bp_dict['children'])):
            node.children.append(cls.from_dict(bp_dict['children'][i]))


class SearchResult:
    def __init__(self, result=None, iserror=False, error_message=""):
        self.result = result
        self.iserror = iserror
        self.error_message = error_message


class BplusTree_t:
    def __init__(self, t):
        self.root = Node()
        self.t = t

    def split_child(self, x, i):
        y = x.children[i]
        z = Node(leaf=y.leaf)

        # Initialize the new node z
        z.n = self.t - 1
        z.keys = y.keys[self.t:]
        z.values = y.values[self.t:]

        y.keys = y.keys[:self.t]
        y.values = y.values[:self.t]

        if not y.leaf:
            z.children = y.children[self.t:]
            y.children = y.children[:self.t]

        # Insert the new node z into x's children list at position i+1
        x.children.insert(i + 1, z)

        # Move the middle key up to the parent node x
        x.keys.insert(i, y.keys[-1])

        # Increment the number of keys in x
        x.n += 1

    def insert_non_full(self, x, k, v):
        i = x.n - 1

        if x.leaf:
            while i >= 0 and k < x.keys[i]:
                print(x.keys(), i)
                x.keys.insert(i + 1, x.keys[i])
                x.values.insert(i + 1, x.values[i])
                i -= 1
            x.keys.insert(i + 1, k)
            x.values.insert(i + 1, v)
            x.n += 1
        else:
            while i >= 0 and k < x.keys[i]:
                i -= 1
            i += 1

            if x.children[i].n == 2 * self.t - 1:
                self.split_child(x, i)
                if k > x.keys[i]:
                    i += 1

            self.insert_non_full(x.children[i], k, v)

    def merge_children(self, x, i):
        y = x.children[i]
        z = x.children[i + 1]

        # Transfer the middle key from x to y
        y.keys.append(x.keys[i])

        # Transfer all keys and children from z to y
        y.keys.extend(z.keys)
        if not y.leaf:
            y.children.extend(z.children)

        # Update the number of keys in y
        y.n += z.n + 1

        # Remove the middle key from x
        x.keys.pop(i)

        # Remove z from x's children list
        x.children.pop(i + 1)

        # Update the number of keys in x
        x.n -= 1

        # Delete the merged node z
        del z

    def redistribute_key(self, x, i):
        y = x.children[i]
        z = x.children[i + 1]

        if y.n < self.t:
            # Move a key from z to y
            y.keys.append(x.keys[i])
            y.n += 1

            # Move a key from z to x
            x.keys[i] = z.keys.pop(0)

            # Move a child from z to y, if applicable
            if not y.leaf:
                y.children.append(z.children.pop(0))

            z.n -= 1
        else:
            # Move a key from y to x
            x.keys[i] = y.keys.pop()

            # Move a key from y to z
            z.keys = y.keys[self.t:] + z.keys
            if not z.leaf:
                z.children = y.children[self.t + 1:] + z.children

            # Update y
            y.keys = y.keys[:self.t]
            y.children = y.children[:self.t + 1]
            y.n = self.t

            # Update z
            z.n = self.t

    def insert(self, k, v):
        if self.root.n == 2 * self.t - 1:
            s = Node(leaf=False)
            s.children.append(self.root)
            self.split_child(s, 0)
            self.insert_non_full(s, k, v)
            self.root = s
        else:
            self.insert_non_full(self.root, k, v)

    def remove(self, k):
        x = self.root
        found = False

        while x is not None:
            i = 0
            while i < x.n and k > x.keys[i]:
                i += 1

            if i < x.n and k == x.keys[i]:
                found = True
                break

            if not x.leaf:
                x = x.children[i]
            else:
                break

        if not found:
            return

        i = 0
        while i < x.n and k > x.keys[i]:
            i += 1
        x.keys.pop(i)
        x.values.pop(i)
        x.n -= 1

        if x.n < self.t and x.parent is not None:
            sibling_index = -1
            sibling = None
            if i != 0:
                sibling_index = i - 1
                sibling = x.parent.children[sibling_index]
            elif i < x.n:
                sibling_index = i
                sibling = x.parent.children[sibling_index]

            if sibling is not None and sibling.n >= self.t:
                self.redistribute_key(x.parent, sibling_index)
            else:
                if sibling_index != -1:
                    self.merge_children(x.parent, sibling_index)
                else:
                    self.merge_children(x.parent, i)

                if x.parent.n == 0:
                    self.root = x
                    del x.parent

    def search(self, k):
        x = self.root
        while not x.leaf:
            i = 0
            while i < x.n and k > x.keys[i]:
                i += 1
            x = x.children[i]  # Continue down the tree until reaching a leaf node

        # Now x is a leaf node, perform the search on this node
        i = 0
        while i < x.n and k > x.keys[i]:
            i += 1
        if i < x.n and k == x.keys[i]:
            return SearchResult(result=x.values[i])
        else:
            return SearchResult(None, True, "Key not found")  # Key not found


class BplusTree:
    def __init__(self, bp_tree: Union[BplusTree_t, object, None] = None):
        if bp_tree is None:
            bp_tree = BplusTree_t(5)
        self.bp_tree = bp_tree
        self.queue = queue.Queue()
        self.event = threading.Event()

        self.thread = threading.Thread(
            target=self._process_loop,
            args=(self.event,)
        )
        self.thread.start()

    def wait(self, timeout=None):
        start_time = time.time()
        while not self.queue.empty():
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError()
            time.sleep(0.02)
        return

    def close(self):
        if self.thread.is_alive():
            self.event.set()
            self.thread.join()

    def _process_loop(self, event: threading.Event):
        while not event.is_set():
            try:
                operate = self.queue.get(timeout=0.3)
                if operate['Operate'] == 'insert':
                    self.bp_tree.insert(operate['key'], operate['value'])

                if operate['Operate'] == 'remove':
                    self.bp_tree.remove(operate['key'])

                self.queue.task_done()

            except queue.Empty:
                pass

    def _submit(self, *args):
        func = inspect.stack()[1].function
        submit_mro = {"Operate": func}
        if func == 'insert':
            submit_mro['key'] = args[0]
            submit_mro['value'] = args[1]

        if func == 'remove':
            submit_mro['key'] = args[0]

        self.queue.put(submit_mro)

    def insert(self, k, v):
        self._submit(k, v)

    def remove(self, k):
        self._submit(k)

    def search(self, k):
        return self.bp_tree.search(k)

    @property
    def root(self):
        return self.bp_tree.root

    def range_query(self, start_key, end_key):
        results = []
        self._range_query_recursive(self.root, start_key, end_key, results)
        return results

    def _range_query_recursive(self, node, start_key, end_key, results):
        if node is None or (not node.leaf and node.keys and start_key < node.keys[0]):
            return

        i = 0
        while i < node.n and node.keys[i] <= end_key:
            if start_key <= node.keys[i]:
                results.append((node.keys[i], node.values[i]))
            i += 1

        # Traverse children if it's an internal node
        if not node.leaf:
            for i in range(node.n + 1):
                self._range_query_recursive(node.children[i], start_key, end_key, results)



import sys
from argparse import ArgumentParser
from pathlib import Path
from rich import print
from natsort import natsorted
import rich.traceback
import rich.color
from random import randint, choice

rich.traceback.install(show_locals=True, max_frames=1)

def _hex(int):
    return hex(int)[2:]

def random_color():
    if randint(0, 1):
        return f"#{_hex(randint(0, 255))}{_hex(randint(0, 255))}{_hex(randint(0, 255))}"
    else:
        return choice(list(rich.color.ANSI_COLOR_NAMES.keys()))

class Loop:
    def __init__(self,iterable, do):
        self.iterable = iterable
        self.do = do
        self.index = 0
        self.value = None

    def exec(self):
        for i, v in enumerate(self.iterable):
            self.index, self.value = i, v
            self.do(self)

    @property
    def item(self):
        return self.index, self.value


class MyPath(Path):
    @property
    def ext(self):
        return ''.join(self.suffixes)[1:]

    @property
    def name_with_no_ext(self):
        return self.name.removesuffix(self.suffix)

    def rename(self, name):
        return super().rename(self.parent / name)


class RenameX:
    def __init__(self, args):
        self.parser = ArgumentParser()
        self.parser.add_argument("path", type=Path,
                            help='批量修改名称的根路径.')
        self.parser.add_argument("--dir", '-d', action="store_true",
                            help='指定Path是一个文件夹,如果Path目标不符合此开关的状态,那么报错(不设置则默认为文件).')
        self.parser.add_argument("--format", '-f', type=str, default='{loop.index}.{file.ext}',
                            help='指定目标名称的格式(使用Python的format格式),默认为"{loop.index}.{file.ext}".')
        self.parser.add_argument('--all', '-a', action='store_true',
                            help='批量修改所有文件,包括文件夹.')

        self.parser.parse_args(args, namespace=self)

        self.loop = None
        self.inners = None

    def __hook(self, loop: Loop):
        i: int ; v: MyPath
        i, v = loop.item

        target = self.format.format(loop=loop, file=v)

        print(
            f'[{random_color()}]Rename [{random_color()}]from [green]"{v}" [{random_color()}]to [green]"{target}" ...', end='')
        if (self.all or v.is_file()) and target != v.name:
            v.rename(target)
            print(f' [{random_color()}]Done.[reset]')
        else:
            print(f' [{random_color()}]Skip.[reset]')

    def exec(self):
        if self.dir and not self.path.is_dir():
            raise ValueError("Don't set '--dir' when path is not a directory")
        if not self.dir and not self.path.is_file():
            raise ValueError("Please set '--dir' when path is a directory")
        if self.dir:
            self.inners = natsorted([MyPath(i) for i in self.path.iterdir()], key=lambda x: x.name)
            # # print('Inners:', self.inners)
            # before = self.inners.copy()
            # self.inners.sort(key=lambda x: x.name_with_no_ext)
            # # print('Inners:', self.inners)
            # # print(before == self.inners)
            self.loop = Loop(self.inners, self.__hook)
            self.loop.exec()


def main():
    main = RenameX(sys.argv[1:])
    main.exec()



import re
from typing import Union

from . import _reconcatenater_abc as _abc

both_set_activate = 0


# 0 raise error
# 1 name only
# 2 ignore only


class REConcater(_abc.REConcater):

    def __init__(self, pattern: Union[_abc.BaseREConcateratable, str]):
        if isinstance(pattern, _abc.BaseREConcateratable):
            self._cur_pattern = pattern
        elif isinstance(pattern, str):
            self._cur_pattern = REConcateratable(pattern)
        else:
            raise TypeError("pattern must be str or BaseREConcateratable")

    @classmethod
    def concat(cls, *args: _abc.BaseREConcateratable):
        return cls(''.join(
            map(lambda x: str(x.pattern()), args)
        ))

    @classmethod
    def or_(cls, *args: _abc.BaseREConcateratable):
        return cls('|'.join(
            map(lambda x: x.pattern(), args)
        ))

    def pattern(self):
        return self._cur_pattern.pattern()

    def match(self, string):
        return self._cur_pattern.match(string)

    def findall(self, string):
        return self._cur_pattern.findall(string)

    def finditer(self, string):
        return self._cur_pattern.finditer(string)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            other = other._cur_pattern
        return self.concat(self._cur_pattern, other)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("can only multiply by int")
        return self.concat(*(self for _ in range(other)))

    def __or__(self, other):
        return self.or_(self._cur_pattern, other)

    def __str__(self):
        return self.pattern()


class BaseREConcateratable(_abc.BaseREConcateratable):
    """
    RE组合基类
    - name 为组命名
    - ignore 是否选择出现在结果列表(groups函数)中，往往出现在需要判断却不需要获取结果的文本部分
    - _pattern 原始文本/RE表达式
    """
    name = None
    ignore = False
    origin_pattern = None

    def pattern(self):

        ignore, name, pattern = self.ignore, self.name, self.origin()  # 提取变量

        if ignore is not False and name is not None:  # ignore=True, name=Any
            if both_set_activate == 0:
                raise NotImplementedError("ignore and name cannot be both set")
            elif both_set_activate == 1:
                ignore = False  # 取消忽略标记
            elif both_set_activate == 2:
                name = None  # 取消命名标记

        if name is not None:  # ignore=False, name=Any
            return "(?P<{name}>{pattern})".format(name=name, pattern=pattern)
        if ignore:  # ignore=True, name=None
            return "(?:{pattern})".format(pattern=pattern)
        else:
            return f"{pattern}"

    def origin(self):
        if isinstance(self.origin_pattern, str):
            return self.origin_pattern
        self.origin_pattern = self.origin_pattern.origin()
        return self.origin()

    def match(self, string):
        return re.match(self.pattern(), string)

    def findall(self, string):
        return re.findall(self.pattern(), string)

    def finditer(self, string):
        return re.finditer(self.pattern(), string)


class REConcateratable(BaseREConcateratable):
    def __init__(self, pattern, name=None, ignore=False):
        if isinstance(pattern, self.__class__):
            self.origin_pattern = pattern.origin()
        self.origin_pattern = pattern
        self.name, self.ignore = name, ignore

    def __add__(self, other):
        return REConcater(self) + other

    def __mul__(self, other):
        return REConcater(self) * other

    def __or__(self, other):
        return REConcater(self) | other

    def __str__(self):
        return self.pattern()


class Re(REConcateratable):
    ...


class Literal(REConcateratable):
    def __init__(self, pattern, name=None, ignore=False):
        super().__init__(
            re.escape(pattern), name, ignore
        )



from abc import abstractmethod, ABC


class PermissonStruct(ABC):
    @abstractmethod
    def __eq__(self, other):
        ...


class PermissionBaseError(BaseException):
    ...


class PermissionObjectExistsError(PermissionBaseError):
    def __init__(self, obj):
        self.error_obj = obj

    def __str__(self):
        return f"Permission object exists(Type: {self.error_obj.__class__.__name__})"


class PermissionNameNotFoundError(PermissionBaseError):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"Permission name not found(Name: {self.name})"


class PermissionSameNameError(PermissionBaseError):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"Permission name is same(Name: {self.name})"


class PermissionSameTypeError(PermissionBaseError):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"Permission name is same(Name: {self.name})"


def merge_roles(*roles):
    role = Role()
    for i in roles:
        role.merge(i)
    return role


class Role:
    def __init__(self, permit_opts=None, prohibit_opts=None):
        if permit_opts is None:
            permit_opts = set()
        if prohibit_opts is None:
            prohibit_opts = set()

        self.permit_opts = set(permit_opts)
        self.prohibit_opts = set(prohibit_opts)

    def _process(self):
        self.permit_opts -= (self.permit_opts & self.prohibit_opts)

    def check(self, opt):
        return opt in self.permit_opts and opt not in self.prohibit_opts

    def prohibit(self, opt):
        self.prohibit_opts.add(opt)
        self._process()

    def permit(self, opt):
        self.prohibit_opts -= {opt}
        self.permit_opts.add(opt)
        self._process()

    def merge(self, other):
        self.permit_opts |= other.permit_opts
        self.prohibit_opts |= other.prohibit_opts
        self._process()


class User(PermissonStruct):
    def __init__(self, name, roles=None):
        if roles is None:
            roles = set()

        self.name = name
        self.roles = set(roles)

        self.groups = set()
        self.domains = set()

        self.last_permission = Role()
        self.update()

    def check(self, opt):
        return self.last_permission.check(opt)

    def add_role(self, role):
        self.roles.add(role)

    def remove_role(self, role):
        self.roles -= {role}

    def join_domain(self, domain):
        self.domains.add(domain)
        domain.add_user(self)

    def leave_domain(self, domain):
        self.domains -= {domain}
        domain.remove_user(self)

    def join_group(self, group):
        self.groups.add(group)
        group.add_user(self)

    def leave_group(self, group):
        self.groups -= {group}
        group.remove_user(self)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name

        return self.name == other

    def update(self):
        last_permission = merge_roles(*self.roles)

        for g in self.groups:
            last_permission.merge(g.last_permission)

        for d in self.domains:
            last_permission.merge(d.last_permission)

        self.last_permission = last_permission


class Group(PermissonStruct):
    def __init__(self, name, roles=None):
        if roles is None:
            roles = set()

        self.name = name
        self.roles = set(roles)

        self.users = set()
        self.domains = set()

        self.last_permission = Role()

    def check(self, opt):
        return self.last_permission.check(opt)

    def add_user(self, user):
        self.users.add(user)

    def add_role(self, role):
        self.roles.add(role)

    def remove_role(self, role):
        self.roles -= {role}

    def remove_user(self, user):
        self.users -= {user}

    def join_domain(self, domain):
        self.domains.add(domain)
        domain.add_group(self)

    def leave_domain(self, domain):
        self.domains -= {domain}
        domain.remove_group(self)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return self.name == other

    def update(self):
        last_permission = merge_roles(*self.roles)

        for d in self.domains:
            last_permission.merge(d.last_permission)

        self.last_permission = last_permission


class Domain(PermissonStruct):
    def __init__(self, name, roles=None):
        if roles is None:
            roles = set()

        self.name = name
        self.roles = set(roles)

        self.users = set()
        self.groups = set()

        self.last_permission = Role()

    def update(self):
        self.last_permission = self.roles

    def check(self, opt):
        return all(role.check(opt) for role in self.roles)

    def add_user(self, user):
        self.users.add(user)

    def add_group(self, group):
        self.groups.add(group)

    def add_role(self, role):
        self.roles.add(role)

    def remove_role(self, role):
        self.roles -= {role}

    def remove_user(self, user):
        self.users -= {user}

    def remove_group(self, group):
        self.groups -= {group}

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return self.name == other


class PermissionManager:
    def __init__(self, roles=None, users=None, groups=None, domains=None):
        if roles is None:
            roles = set()
        if users is None:
            users = set()
        if groups is None:
            groups = set()
        if domains is None:
            domains = set()
        self.objects = {}  # 使用字典存储对象
        for u in users:
            self.objects[u.name] = u
        for g in groups:
            self.objects[g.name] = g
        for d in domains:
            self.objects[d.name] = d

        self.default_roles = roles

    def create_user(self, name):
        if name in self.objects:
            raise PermissionObjectExistsError(self.objects[name])
        u = User(name, self.default_roles)
        self.objects[name] = u

    def create_group(self, name):
        if name in self.objects:
            raise PermissionObjectExistsError(self.objects[name])
        g = Group(name, self.default_roles)
        self.objects[name] = g
        return g

    def create_domain(self, name):
        if name in self.objects:
            raise PermissionObjectExistsError(self.objects[name])
        d = Domain(name, self.default_roles)
        self.objects[name] = d
        return d

    def exists(self, name):
        return name in self.objects

    def remove(self, name):
        if name not in self.objects:
            raise Exception("permission object not exists")
        obj = self.objects.pop(name)  # type: PermissonStruct | User | Group | Domain
        if isinstance(obj, User):
            for group in obj.groups:
                group.remove_user(obj)

            for domain in obj.domains:
                domain.remove_user(obj)

        if isinstance(obj, Group):
            for user in obj.users:
                user.groups -= {obj}

            for domain in obj.domains:
                domain.leave_group(obj)

        if isinstance(obj, Domain):
            for user in obj.users:
                user.domains -= {obj}

            for group in obj.groups:
                group.remove_domain(obj)

    def update(self):
        domains = [i for i in self.objects.values() if isinstance(i, Domain)]
        for d in domains:
            d.update()

        groups = [i for i in self.objects.values() if isinstance(i, Group)]
        for g in groups:
            g.update()

        users = [i for i in self.objects.values() if isinstance(i, User)]
        for u in users:
            u.update()

    NonTopTypes = (Group, User)
    ContainerTypes = (Group, Domain)

    def let_join(self, obj: NonTopTypes, container: ContainerTypes):
        if type(obj) is type(container):  # type check
            raise Exception("can not join same type")

        if type(obj) is User:
            if type(container) is Group:
                container.add_user(obj)
                obj.join_group(container)

            if type(container) is Domain:
                container.add_user(obj)
                obj.join_domain(container)
        elif type(obj) is Group:
            if type(container) is Domain:
                container.add_group(obj)
                obj.join_domain(container)

    def let_leave(self, obj: NonTopTypes, container: ContainerTypes):
        if type(obj) is type(container):  # type check
            raise Exception("can not leave same type")
        if type(obj) is User:
            if type(container) is Group:
                container.remove_user(obj)
                obj.leave_group(container)

            if type(container) is Domain:
                container.remove_user(obj)
                obj.leave_domain(container)

        elif type(obj) is Group:
            if type(container) is Domain:
                container.remove_group(obj)
                obj.leave_domain(container)

    def get_by_name(self, name) -> User | Group | Domain | None:
        return self.objects.get(name)

    def isinstance(self, name, _type):
        obj = self.objects.get(name)
        return isinstance(obj, _type) if obj else False

    def is_group(self, name):
        return self.isinstance(self.objects[name], Group)

    def is_user(self, name):
        return self.isinstance(self.objects[name], User)

    def is_domain(self, name):
        return self.isinstance(self.objects[name], Domain)

    def check(self, target_name, opt):
        if target_name not in self.objects:
            raise PermissionNameNotFoundError(target_name)

        target = self.objects[target_name]
        return target.check(opt)


if __name__ == '__main__':
    # Test for PermissionManager
    pm = PermissionManager()
    pm.create_user("user1")
    pm.create_user("user2")
    pm.create_user("user3")

    pm.create_group("group1")
    pm.create_group("group2")
    pm.create_group("group3")

    print(pm.objects)

    usr1 = pm.get_by_name("user1")
    print(usr1)
    usr1.add_role(Role({'R', 'W'}, {'X', 'Y'}))
    pm.get_by_name("user2").add_role(Role({'R'}, {'X'}))
    pm.get_by_name("user3").add_role(Role({'R', 'W', 'X', 'Y'}, ))

    pm.get_by_name("group1").add_role(Role({'R'}, {'X'}))
    pm.get_by_name("group2").add_role(Role({'R'}, {'Y'}))
    pm.get_by_name("group3").add_role(Role({'R', 'X'}, {'W'}))

    pm.update()

    print(pm.check("user1", "R"))
    print(pm.check("user1", "X"))
    print(pm.check("user1", "W"))
    print(pm.is_user("user1"))
    print(pm.is_user("group1"))

    pm.let_join(pm.get_by_name("user1"), pm.get_by_name("group1"))
    pm.let_join(pm.get_by_name("user2"), pm.get_by_name("group2"))

    pm.update()

    print(pm.check("user1", "R"))
    print(pm.check("user1", "X"))



import asyncio
import socket
from typing import *

from .._methods import get_part_from_sock, build_part_into_sock
from ...._hycore.socket import Asyncsocket
from ....hystruct.Serializers import Serializer, dumps, loads


class serialized_socket:
    def __init__(self, s: Union[socket.socket, Any] = None, loop: asyncio.AbstractEventLoop = None,
                 serializer: Serializer = dumps):
        self.loop = loop if loop else asyncio.get_running_loop()
        self.serializer = serializer
        self.s = Asyncsocket(s)
        self._forever_stop = False

    async def connect(self, addr, port, timeout=None):
        await self.s.connect((addr, port), timeout)

    async def send(self, data):
        bytes_data = dumps(data, serializer=self.serializer)
        await build_part_into_sock(bytes_data, self.s)

    async def send_iter(self, items):
        for item in items:
            await self.send(item)

    async def send_aiter(self, items):
        async for item in items:
            await self.send(item)

    async def recv(self):
        bytes_data = await get_part_from_sock(self.s)
        return loads(bytes_data)

    async def recv_iter(self, size=1):
        for i in range(size):
            yield await self.recv()

    async def recv_forever(self):
        while not self._forever_stop:
            yield await self.recv()

    async def close(self):
        self._forever_stop = True
        await self.s.close()



import json.decoder

from .file import empty, isspace
from .json import Json, Pickle


class JsonDict(dict):
    def __init__(self, file):
        super().__init__()
        self.file = file
        self.update(Json.load(file))

    def save(self):
        Json.dump(self, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


class JsonList(list):
    def __init__(self, file):
        super().__init__()
        self.file = file
        self.extend(Json.load(file))

    def save(self):
        Json.dump(self, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


class JsonBaseType:
    def __init__(self, file):
        self.file = file
        self.data = Json.load(file)
        if not isinstance(
                self.data, int | float | str
        ):
            raise TypeError(
                f"{self.__class__.__name__} can only be used with base types"
            )

    def save(self):
        Json.dump(self.data, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


class JsonPickleDict(dict):
    def __init__(self, file):
        super().__init__()
        self.file = file
        self.update(Pickle.load(file))

    def save(self):
        Pickle.dump(self, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


class JsonPickleList(list):
    def __init__(self, file):
        super().__init__()
        self.file = file
        self.extend(Pickle.load(file))

    def save(self):
        Pickle.dump(self, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


class JsonPickleTypes:
    def __init__(self, file):
        self.file = file
        self.errored = False
        try:
            self.obj = Pickle.load(file)
        except (json.decoder.JSONDecodeError, Pickle.jsonpickle.unpickler.errors.ClassNotFoundError):
            self.obj = None
            self.errored = True

    def save(self):
        Pickle.dump(self.obj, self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


def json_open(file):
    dtype = type(Json.load(file))
    if empty(file) or isspace(file):
        return None
    if dtype == dict:
        return JsonDict(file)
    elif dtype == list:
        return JsonList(file)
    elif dtype == int or dtype == float or dtype == str:
        return JsonBaseType(file)
    else:
        raise TypeError(f"Unknown type: {dtype}")


def pickle_open(file):
    if empty(file) or isspace(file):
        return None
    dtype = type(Pickle.load(file))
    if dtype == dict:
        return JsonPickleDict(file)
    elif dtype == list:
        return JsonPickleList(file)
    else:
        return JsonPickleTypes(file)


def json_safe_open(file, dtype=dict):
    if empty(file) or isspace(file):
        Json.dump(dtype(), file)
    return json_open(file)


def pickle_safe_open(file, dtype):
    if empty(file) or isspace(file):
        Pickle.dump(dtype(), file)

    return pickle_open(file)


def json_read(file):
    return Json.load(file)


def json_write(data, file):
    return Json.dump(data, file)


def pickle_read(file):
    return Pickle.load(file)


def pickle_write(data, file):
    return Pickle.dump(data, file)
